"""Add 'displayed_by_user' column to ArchivedMessage, for xmpp.MARK_ALL_MESSAGES

Revision ID: 86d980dbb426
Revises: 259a1b71eadf
Create Date: 2025-12-02 20:47:51.919688

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "86d980dbb426"
down_revision: str | None = "259a1b71eadf"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table("mam", schema=None) as batch_op:
        batch_op.add_column(sa.Column("displayed_by_user", sa.Boolean()))
    # Set all existing rows to displayed, to avoid sending large busts of read markers
    # for messages archived before the introduction of the new column.
    mam = sa.table("mam", sa.column("displayed_by_user", sa.Boolean()))
    op.execute(mam.update().values(displayed_by_user=True))


def downgrade() -> None:
    with op.batch_alter_table("mam", schema=None) as batch_op:
        batch_op.drop_column("displayed_by_user")



from __future__ import annotations

import hashlib
import logging
import shutil
import uuid
from collections.abc import Collection, Iterator
from datetime import UTC, datetime, timedelta
from mimetypes import guess_extension

import sqlalchemy as sa
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0231.stanza import BitsOfBinary
from sqlalchemy import Engine, delete, event, select, update
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.orm import Session, attributes, sessionmaker

from ..core import config
from ..util.archive_msg import HistoryMessage
from ..util.types import MamMetadata, Sticker
from .meta import Base
from .models import (
    ArchivedMessage,
    ArchivedMessageSource,
    Avatar,
    Bob,
    Contact,
    ContactSent,
    DirectMessages,
    DirectThreads,
    GatewayUser,
    GroupMessages,
    GroupMessagesOrigin,
    GroupThreads,
    Participant,
    Room,
)


class UpdatedMixin:
    model: type[Base] = NotImplemented

    def __init__(self, session: Session) -> None:
        self.reset_updated(session)

    def get_by_pk(self, session: Session, pk: int) -> type[Base]:
        stmt = select(self.model).where(self.model.id == pk)  # type:ignore
        return session.scalar(stmt)

    def reset_updated(self, session: Session) -> None:
        session.execute(update(self.model).values(updated=False))


class SlidgeStore:
    def __init__(self, engine: Engine) -> None:
        self._engine = engine
        self.session = sessionmaker(engine)

        self.users = UserStore(self.session)
        self.avatars = AvatarStore(self.session)
        self.id_map = IdMapStore()
        self.bob = BobStore()
        with self.session() as session:
            self.contacts = ContactStore(session)
            self.mam = MAMStore(session, self.session)
            self.rooms = RoomStore(session)
            self.participants = ParticipantStore(session)
            session.commit()


class UserStore:
    def __init__(self, session_maker) -> None:
        self.session = session_maker

    def update(self, user: GatewayUser) -> None:
        with self.session(expire_on_commit=False) as session:
            # https://github.com/sqlalchemy/sqlalchemy/discussions/6473
            try:
                attributes.flag_modified(user, "legacy_module_data")
                attributes.flag_modified(user, "preferences")
            except InvalidRequestError:
                pass
            session.add(user)
            session.commit()


class AvatarStore:
    def __init__(self, session_maker) -> None:
        self.session = session_maker


LegacyToXmppType = (
    type[DirectMessages]
    | type[DirectThreads]
    | type[GroupMessages]
    | type[GroupThreads]
    | type[GroupMessagesOrigin]
)


class IdMapStore:
    @staticmethod
    def _set(
        session: Session,
        foreign_key: int,
        legacy_id: str,
        xmpp_ids: list[str],
        type_: LegacyToXmppType,
    ) -> None:
        kwargs = dict(foreign_key=foreign_key, legacy_id=legacy_id)
        ids = session.scalars(
            select(type_.id).filter(
                type_.foreign_key == foreign_key, type_.legacy_id == legacy_id
            )
        )
        if ids:
            log.debug("Resetting legacy ID %s", legacy_id)
        session.execute(delete(type_).where(type_.id.in_(ids)))
        for xmpp_id in xmpp_ids:
            msg = type_(xmpp_id=xmpp_id, **kwargs)
            session.add(msg)

    def set_thread(
        self,
        session: Session,
        foreign_key: int,
        legacy_id: str,
        xmpp_id: str,
        group: bool,
    ) -> None:
        self._set(
            session,
            foreign_key,
            legacy_id,
            [xmpp_id],
            GroupThreads if group else DirectThreads,
        )

    def set_msg(
        self,
        session: Session,
        foreign_key: int,
        legacy_id: str,
        xmpp_ids: list[str],
        group: bool,
    ) -> None:
        self._set(
            session,
            foreign_key,
            legacy_id,
            xmpp_ids,
            GroupMessages if group else DirectMessages,
        )

    def set_origin(
        self, session: Session, foreign_key: int, legacy_id: str, xmpp_id: str
    ) -> None:
        self._set(
            session,
            foreign_key,
            legacy_id,
            [xmpp_id],
            GroupMessagesOrigin,
        )

    def get_origin(
        self, session: Session, foreign_key: int, legacy_id: str
    ) -> list[str]:
        return self._get(
            session,
            foreign_key,
            legacy_id,
            GroupMessagesOrigin,
        )

    @staticmethod
    def _get(
        session: Session, foreign_key: int, legacy_id: str, type_: LegacyToXmppType
    ) -> list[str]:
        return list(
            session.scalars(
                select(type_.xmpp_id).filter_by(
                    foreign_key=foreign_key, legacy_id=str(legacy_id)
                )
            )
        )

    def get_xmpp(
        self, session: Session, foreign_key: int, legacy_id: str, group: bool
    ) -> list[str]:
        return self._get(
            session,
            foreign_key,
            legacy_id,
            GroupMessages if group else DirectMessages,
        )

    @staticmethod
    def _get_legacy(
        session: Session, foreign_key: int, xmpp_id: str, type_: LegacyToXmppType
    ) -> str | None:
        return session.scalar(
            select(type_.legacy_id).filter_by(foreign_key=foreign_key, xmpp_id=xmpp_id)
        )

    def get_legacy(
        self,
        session: Session,
        foreign_key: int,
        xmpp_id: str,
        group: bool,
        origin: bool = False,
    ) -> str | None:
        if origin and group:
            return self._get_legacy(
                session,
                foreign_key,
                xmpp_id,
                GroupMessagesOrigin,
            )
        return self._get_legacy(
            session,
            foreign_key,
            xmpp_id,
            GroupMessages if group else DirectMessages,
        )

    def get_thread(
        self, session: Session, foreign_key: int, xmpp_id: str, group: bool
    ) -> str | None:
        return self._get_legacy(
            session,
            foreign_key,
            xmpp_id,
            GroupThreads if group else DirectThreads,
        )

    @staticmethod
    def was_sent_by_user(
        session: Session, foreign_key: int, legacy_id: str, group: bool
    ) -> bool:
        type_ = GroupMessages if group else DirectMessages
        return (
            session.scalar(
                select(type_.id).filter_by(foreign_key=foreign_key, legacy_id=legacy_id)
            )
            is not None
        )


class ContactStore(UpdatedMixin):
    model = Contact

    def __init__(self, session: Session) -> None:
        super().__init__(session)
        session.execute(update(Contact).values(cached_presence=False))
        session.execute(update(Contact).values(caps_ver=None))

    @staticmethod
    def add_to_sent(session: Session, contact_pk: int, msg_id: str) -> None:
        if (
            session.query(ContactSent.id)
            .where(ContactSent.contact_id == contact_pk)
            .where(ContactSent.msg_id == msg_id)
            .first()
        ) is not None:
            log.warning("Contact %s has already sent message %s", contact_pk, msg_id)
            return
        new = ContactSent(contact_id=contact_pk, msg_id=msg_id)
        session.add(new)

    @staticmethod
    def pop_sent_up_to(session: Session, contact_pk: int, msg_id: str) -> list[str]:
        result = []
        to_del = []
        for row in session.execute(
            select(ContactSent)
            .where(ContactSent.contact_id == contact_pk)
            .order_by(ContactSent.id)
        ).scalars():
            to_del.append(row.id)
            result.append(row.msg_id)
            if row.msg_id == msg_id:
                break
        session.execute(delete(ContactSent).where(ContactSent.id.in_(to_del)))
        return result


class MAMStore:
    def __init__(self, session: Session, session_maker) -> None:
        self.session = session_maker
        self.reset_source(session)

    @staticmethod
    def reset_source(session: Session) -> None:
        session.execute(
            update(ArchivedMessage).values(source=ArchivedMessageSource.BACKFILL)
        )

    @staticmethod
    def nuke_older_than(session: Session, days: int) -> None:
        session.execute(
            delete(ArchivedMessage).where(
                ArchivedMessage.timestamp < datetime.now() - timedelta(days=days)
            )
        )

    @staticmethod
    def add_message(
        session: Session,
        room_pk: int,
        message: HistoryMessage,
        archive_only: bool,
        legacy_msg_id: str | None,
    ) -> None:
        source = (
            ArchivedMessageSource.BACKFILL
            if archive_only
            else ArchivedMessageSource.LIVE
        )
        existing = session.execute(
            select(ArchivedMessage)
            .where(ArchivedMessage.room_id == room_pk)
            .where(ArchivedMessage.stanza_id == message.id)
        ).scalar()
        if existing is None and legacy_msg_id is not None:
            existing = session.execute(
                select(ArchivedMessage)
                .where(ArchivedMessage.room_id == room_pk)
                .where(ArchivedMessage.legacy_id == str(legacy_msg_id))
            ).scalar()
        if existing is not None:
            log.debug("Updating message %s in room %s", message.id, room_pk)
            existing.timestamp = message.when
            existing.stanza = str(message.stanza)
            existing.author_jid = message.stanza.get_from()
            existing.source = source
            existing.legacy_id = legacy_msg_id
            session.add(existing)
            return
        mam_msg = ArchivedMessage(
            stanza_id=message.id,
            timestamp=message.when,
            stanza=str(message.stanza),
            author_jid=message.stanza.get_from(),
            room_id=room_pk,
            source=source,
            legacy_id=legacy_msg_id,
        )
        session.add(mam_msg)

    @staticmethod
    def get_messages(
        session: Session,
        room_pk: int,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        before_id: str | None = None,
        after_id: str | None = None,
        ids: Collection[str] = (),
        last_page_n: int | None = None,
        sender: str | None = None,
        flip: bool = False,
    ) -> Iterator[HistoryMessage]:
        q = select(ArchivedMessage).where(ArchivedMessage.room_id == room_pk)
        if start_date is not None:
            q = q.where(ArchivedMessage.timestamp >= start_date)
        if end_date is not None:
            q = q.where(ArchivedMessage.timestamp <= end_date)
        if before_id is not None:
            stamp = session.execute(
                select(ArchivedMessage.timestamp).where(
                    ArchivedMessage.stanza_id == before_id,
                    ArchivedMessage.room_id == room_pk,
                )
            ).scalar_one_or_none()
            if stamp is None:
                raise XMPPError(
                    "item-not-found",
                    f"Message {before_id} not found",
                )
            q = q.where(ArchivedMessage.timestamp < stamp)
        if after_id is not None:
            stamp = session.execute(
                select(ArchivedMessage.timestamp).where(
                    ArchivedMessage.stanza_id == after_id,
                    ArchivedMessage.room_id == room_pk,
                )
            ).scalar_one_or_none()
            if stamp is None:
                raise XMPPError(
                    "item-not-found",
                    f"Message {after_id} not found",
                )
            q = q.where(ArchivedMessage.timestamp > stamp)
        if ids:
            q = q.filter(ArchivedMessage.stanza_id.in_(ids))
        if sender is not None:
            q = q.where(ArchivedMessage.author_jid == sender)
        if flip:
            q = q.order_by(ArchivedMessage.timestamp.desc())
        else:
            q = q.order_by(ArchivedMessage.timestamp.asc())
        msgs = list(session.execute(q).scalars())
        if ids and len(msgs) != len(ids):
            raise XMPPError(
                "item-not-found",
                "One of the requested messages IDs could not be found "
                "with the given constraints.",
            )
        if last_page_n is not None:
            if flip:
                msgs = msgs[:last_page_n]
            else:
                msgs = msgs[-last_page_n:]
        for h in msgs:
            yield HistoryMessage(
                stanza=str(h.stanza), when=h.timestamp.replace(tzinfo=UTC)
            )

    @staticmethod
    def get_first(
        session: Session, room_pk: int, with_legacy_id: bool = False
    ) -> ArchivedMessage | None:
        q = (
            select(ArchivedMessage)
            .where(ArchivedMessage.room_id == room_pk)
            .order_by(ArchivedMessage.timestamp.asc())
        )
        if with_legacy_id:
            q = q.filter(ArchivedMessage.legacy_id.isnot(None))
        return session.execute(q).scalar()

    @staticmethod
    def get_last(
        session: Session, room_pk: int, source: ArchivedMessageSource | None = None
    ) -> ArchivedMessage | None:
        q = select(ArchivedMessage).where(ArchivedMessage.room_id == room_pk)

        if source is not None:
            q = q.where(ArchivedMessage.source == source)

        return session.execute(q.order_by(ArchivedMessage.timestamp.desc())).scalar()

    def get_first_and_last(self, session: Session, room_pk: int) -> list[MamMetadata]:
        r = []
        first = self.get_first(session, room_pk)
        if first is not None:
            r.append(MamMetadata(first.stanza_id, first.timestamp))
        last = self.get_last(session, room_pk)
        if last is not None:
            r.append(MamMetadata(last.stanza_id, last.timestamp))
        return r

    @staticmethod
    def get_most_recent_with_legacy_id(
        session: Session, room_pk: int, source: ArchivedMessageSource | None = None
    ) -> ArchivedMessage | None:
        q = (
            select(ArchivedMessage)
            .where(ArchivedMessage.room_id == room_pk)
            .where(ArchivedMessage.legacy_id.isnot(None))
        )
        if source is not None:
            q = q.where(ArchivedMessage.source == source)
        return session.execute(q.order_by(ArchivedMessage.timestamp.desc())).scalar()

    @staticmethod
    def get_least_recent_with_legacy_id_after(
        session: Session,
        room_pk: int,
        after_id: str,
        source: ArchivedMessageSource = ArchivedMessageSource.LIVE,
    ) -> ArchivedMessage | None:
        after_timestamp = (
            session.query(ArchivedMessage.timestamp)
            .filter(ArchivedMessage.room_id == room_pk)
            .filter(ArchivedMessage.legacy_id == after_id)
            .scalar()
        )
        q = (
            select(ArchivedMessage)
            .where(ArchivedMessage.room_id == room_pk)
            .where(ArchivedMessage.legacy_id.isnot(None))
            .where(ArchivedMessage.source == source)
            .where(ArchivedMessage.timestamp > after_timestamp)
        )
        return session.execute(q.order_by(ArchivedMessage.timestamp.asc())).scalar()

    @staticmethod
    def get_by_legacy_id(
        session: Session, room_pk: int, legacy_id: str
    ) -> ArchivedMessage | None:
        return (
            session.query(ArchivedMessage)
            .filter(ArchivedMessage.room_id == room_pk)
            .filter(ArchivedMessage.legacy_id == legacy_id)
            .first()
        )

    @staticmethod
    def pop_unread_up_to(session: Session, room_pk: int, stanza_id: str) -> list[str]:
        q = (
            select(ArchivedMessage.id, ArchivedMessage.stanza_id)
            .where(ArchivedMessage.room_id == room_pk)
            .where(~ArchivedMessage.displayed_by_user)
            .where(ArchivedMessage.legacy_id.is_not(None))
            .order_by(ArchivedMessage.timestamp.asc())
        )

        ref = session.scalar(
            select(ArchivedMessage)
            .where(ArchivedMessage.room_id == room_pk)
            .where(ArchivedMessage.stanza_id == stanza_id)
        )

        if ref is None:
            log.debug(
                "(pop unread in muc): message not found, returning all MAM messages."
            )
            rows = session.execute(q)
        else:
            rows = session.execute(q.where(ArchivedMessage.timestamp <= ref.timestamp))

        pks: list[int] = []
        stanza_ids: list[str] = []

        for id_, stanza_id in rows:
            pks.append(id_)
            stanza_ids.append(stanza_id)

        session.execute(
            update(ArchivedMessage)
            .where(ArchivedMessage.id.in_(pks))
            .values(displayed_by_user=True)
        )
        return stanza_ids

    @staticmethod
    def is_displayed_by_user(
        session: Session, room_jid: str, legacy_msg_id: str
    ) -> bool:
        return any(
            session.execute(
                select(ArchivedMessage.displayed_by_user)
                .join(Room)
                .where(Room.jid == room_jid)
                .where(ArchivedMessage.legacy_id == legacy_msg_id)
            ).scalars()
        )


class RoomStore(UpdatedMixin):
    model = Room

    def reset_updated(self, session: Session) -> None:
        super().reset_updated(session)
        session.execute(
            update(Room).values(
                subject_setter=None,
                user_resources=None,
                history_filled=False,
                participants_filled=False,
            )
        )

    @staticmethod
    def get_all(session: Session, user_pk: int) -> Iterator[Room]:
        yield from session.scalars(select(Room).where(Room.user_account_id == user_pk))

    @staticmethod
    def get(session: Session, user_pk: int, legacy_id: str) -> Room:
        return session.execute(
            select(Room)
            .where(Room.user_account_id == user_pk)
            .where(Room.legacy_id == legacy_id)
        ).scalar_one()

    @staticmethod
    def nick_available(session: Session, room_pk: int, nickname: str) -> bool:
        return (
            session.execute(
                select(Participant.id).filter_by(room_id=room_pk, nickname=nickname)
            )
        ).one_or_none() is None


class ParticipantStore:
    def __init__(self, session: Session) -> None:
        session.execute(delete(Participant))

    @staticmethod
    def get_all(
        session: Session, room_pk: int, user_included: bool = True
    ) -> Iterator[Participant]:
        query = select(Participant).where(Participant.room_id == room_pk)
        if not user_included:
            query = query.where(~Participant.is_user)
        yield from session.scalars(query).unique()


class BobStore:
    _ATTR_MAP = {
        "sha-1": "sha_1",
        "sha1": "sha_1",
        "sha-256": "sha_256",
        "sha256": "sha_256",
        "sha-512": "sha_512",
        "sha512": "sha_512",
    }

    _ALG_MAP = {
        "sha_1": hashlib.sha1,
        "sha_256": hashlib.sha256,
        "sha_512": hashlib.sha512,
    }

    def __init__(self) -> None:
        if (config.HOME_DIR / "slidge_stickers").exists():
            shutil.move(
                config.HOME_DIR / "slidge_stickers", config.HOME_DIR / "bob_store"
            )
        self.root_dir = config.HOME_DIR / "bob_store"
        self.root_dir.mkdir(exist_ok=True)

    @staticmethod
    def __split_cid(cid: str) -> list[str]:
        return cid.removesuffix("@bob.xmpp.org").split("+")

    def __get_condition(self, cid: str):
        alg_name, digest = self.__split_cid(cid)
        attr = self._ATTR_MAP.get(alg_name)
        if attr is None:
            log.warning("Unknown hash algorithm: %s", alg_name)
            return None
        return getattr(Bob, attr) == digest

    def get(self, session: Session, cid: str) -> Bob | None:
        try:
            return session.query(Bob).filter(self.__get_condition(cid)).scalar()
        except ValueError:
            log.warning("Cannot get Bob with CID: %s", cid)
            return None

    def get_sticker(self, session: Session, cid: str) -> Sticker | None:
        bob = self.get(session, cid)
        if bob is None:
            return None
        return Sticker(
            self.root_dir / bob.file_name,
            bob.content_type,
            {h: getattr(bob, h) for h in self._ALG_MAP},
        )

    def get_bob(
        self, session: Session, _jid, _node, _ifrom, cid: str
    ) -> BitsOfBinary | None:
        stored = self.get(session, cid)
        if stored is None:
            return None
        bob = BitsOfBinary()
        bob["data"] = (self.root_dir / stored.file_name).read_bytes()
        if stored.content_type is not None:
            bob["type"] = stored.content_type
        bob["cid"] = cid
        return bob

    def del_bob(self, session: Session, _jid, _node, _ifrom, cid: str) -> None:
        try:
            file_name = session.scalar(
                delete(Bob).where(self.__get_condition(cid)).returning(Bob.file_name)
            )
        except ValueError:
            log.warning("Cannot delete Bob with CID: %s", cid)
            return None
        if file_name is None:
            log.warning("No BoB with CID: %s", cid)
            return None
        (self.root_dir / file_name).unlink()

    def set_bob(self, session: Session, _jid, _node, _ifrom, bob: BitsOfBinary) -> None:
        cid = bob["cid"]
        try:
            alg_name, digest = self.__split_cid(cid)
        except ValueError:
            log.warning("Invalid CID provided: %s", cid)
            return
        attr = self._ATTR_MAP.get(alg_name)
        if attr is None:
            log.warning("Cannot set Bob: Unknown algorithm type: %s", alg_name)
            return
        existing = self.get(session, bob["cid"])
        if existing:
            log.debug("Bob already exists")
            return
        bytes_ = bob["data"]
        path = self.root_dir / uuid.uuid4().hex
        if bob["type"]:
            path = path.with_suffix(guess_extension(bob["type"]) or "")
        path.write_bytes(bytes_)
        hashes = {k: v(bytes_).hexdigest() for k, v in self._ALG_MAP.items()}
        if hashes[attr] != digest:
            path.unlink(missing_ok=True)
            raise ValueError("Provided CID does not match calculated hash")
        row = Bob(file_name=path.name, content_type=bob["type"] or None, **hashes)
        session.add(row)


@event.listens_for(sa.orm.Session, "after_flush")
def _check_avatar_orphans(session, flush_context):
    if not session.deleted:
        return

    potentially_orphaned = set()
    for obj in session.deleted:
        if isinstance(obj, (Contact, Room)) and obj.avatar_id:
            potentially_orphaned.add(obj.avatar_id)
    if not potentially_orphaned:
        return

    result = session.execute(
        sa.delete(Avatar).where(
            sa.and_(
                Avatar.id.in_(potentially_orphaned),
                sa.not_(sa.exists().where(Contact.avatar_id == Avatar.id)),
                sa.not_(sa.exists().where(Room.avatar_id == Avatar.id)),
            )
        )
    )
    deleted_count = result.rowcount
    log.debug("Auto-deleted %s orphaned avatars", deleted_count)


log = logging.getLogger(__name__)



import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from slixmpp import (
    JID,
    CoroutineCallback,
    Iq,
    Presence,
    StanzaPath,
    register_stanza_plugin,
)
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.plugins.base import BasePlugin, register_plugin
from slixmpp.plugins.xep_0060.stanza import Event, EventItem, EventItems, Item
from slixmpp.plugins.xep_0084 import Data as AvatarData
from slixmpp.plugins.xep_0084 import MetaData as AvatarMetadata
from slixmpp.plugins.xep_0172 import UserNick
from slixmpp.plugins.xep_0292.stanza import VCard4
from slixmpp.types import JidStr, OptJidStr

from ..db.avatar import CachedAvatar, avatar_cache
from ..db.models import GatewayUser
from ..util.lock import NamedLockMixin

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway

    from ..contact.contact import LegacyContact

VCARD4_NAMESPACE = "urn:xmpp:vcard4"


class PepAvatar:
    def __init__(self) -> None:
        self.metadata: AvatarMetadata | None = None
        self.id: str | None = None
        self._avatar_data_path: Path | None = None

    @property
    def data(self) -> AvatarData | None:
        if self._avatar_data_path is None:
            return None
        data = AvatarData()
        data.set_value(self._avatar_data_path.read_bytes())
        return data

    def set_avatar_from_cache(self, cached_avatar: CachedAvatar) -> None:
        metadata = AvatarMetadata()
        self.id = cached_avatar.hash
        metadata.add_info(
            id=cached_avatar.hash,
            itype="image/png",
            ibytes=cached_avatar.path.stat().st_size,
            height=str(cached_avatar.height),
            width=str(cached_avatar.width),
        )
        self.metadata = metadata
        self._avatar_data_path = cached_avatar.path


class PubSubComponent(NamedLockMixin, BasePlugin):
    xmpp: "BaseGateway"

    name = "pubsub"
    description = "Pubsub component"
    dependencies = {
        "xep_0030",
        "xep_0060",
        "xep_0115",
        "xep_0163",
    }
    default_config = {"component_name": None}
    component_name: str

    def __init__(self, *a, **kw) -> None:
        super().__init__(*a, **kw)
        register_stanza_plugin(EventItem, UserNick)

    def plugin_init(self) -> None:
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_avatar_data",
                StanzaPath(f"iq@type=get/pubsub/items@node={AvatarData.namespace}"),
                self._get_avatar_data,  # type:ignore
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_avatar_metadata",
                StanzaPath(f"iq@type=get/pubsub/items@node={AvatarMetadata.namespace}"),
                self._get_avatar_metadata,  # type:ignore
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_vcard",
                StanzaPath(f"iq@type=get/pubsub/items@node={VCARD4_NAMESPACE}"),
                self._get_vcard,  # type:ignore
            )
        )

        disco = self.xmpp.plugin["xep_0030"]
        disco.add_identity("pubsub", "pep", self.component_name)
        disco.add_identity("account", "registered", self.component_name)
        disco.add_feature("http://jabber.org/protocol/pubsub#event")
        disco.add_feature("http://jabber.org/protocol/pubsub#retrieve-items")
        disco.add_feature("http://jabber.org/protocol/pubsub#persistent-items")

    async def __get_features(self, presence: Presence) -> list[str]:
        from_ = presence.get_from()
        ver_string = presence["caps"]["ver"]
        if ver_string:
            info = await self.xmpp.plugin["xep_0115"].get_caps(from_)
        else:
            info = None
        if info is None:
            async with self.lock(from_):
                try:
                    iq = await self.xmpp.plugin["xep_0030"].get_info(from_)
                except (IqError, IqTimeout):
                    log.debug("Could not get disco#info of %s, ignoring", from_)
                    return []
            info = iq["disco_info"]
        return info["features"]

    async def on_presence_available(
        self, p: Presence, contact: Optional["LegacyContact"]
    ) -> None:
        if p.get_plugin("muc_join", check=True) is not None:
            log.debug("Ignoring MUC presence here")
            return

        to = p.get_to()
        if to != self.xmpp.boundjid.bare:
            # we don't want to push anything for contacts that are not in the user's roster
            if contact is None or not contact.is_friend:
                return

        from_ = p.get_from()
        features = await self.__get_features(p)

        if AvatarMetadata.namespace + "+notify" in features:
            try:
                pep_avatar = await self._get_authorized_avatar(p, contact)
            except XMPPError:
                pass
            else:
                if pep_avatar.metadata is not None:
                    await self.__broadcast(
                        data=pep_avatar.metadata,
                        from_=p.get_to().bare,
                        to=from_,
                        id=pep_avatar.metadata["info"]["id"],
                    )
        if UserNick.namespace + "+notify" in features:
            try:
                pep_nick = await self._get_authorized_nick(p, contact)
            except XMPPError:
                pass
            else:
                await self.__broadcast(data=pep_nick, from_=p.get_to(), to=from_)

        if contact is not None and VCARD4_NAMESPACE + "+notify" in features:
            await self.broadcast_vcard_event(
                p.get_to(), from_, await contact.get_vcard()
            )

    async def broadcast_vcard_event(
        self, from_: JID, to: JID, vcard: VCard4 | None
    ) -> None:
        item = Item()
        item.namespace = VCARD4_NAMESPACE
        item["id"] = "current"
        # vcard: VCard4 = await self.xmpp["xep_0292_provider"].get_vcard(from_, to)
        # The vcard content should NOT be in this event according to the spec:
        # https://xmpp.org/extensions/xep-0292.html#sect-idm45669698174224
        # but movim expects it to be here, and I guess it does not hurt

        log.debug("Broadcast vcard4 event: %s", vcard)
        await self.__broadcast(
            data=vcard,
            from_=JID(from_).bare,
            to=to,
            id="current",
            node=VCARD4_NAMESPACE,
        )

    async def __get_contact(self, stanza: Iq | Presence):
        session = self.xmpp.get_session_from_stanza(stanza)
        return await session.contacts.by_jid(stanza.get_to())

    async def _get_authorized_avatar(
        self, stanza: Iq | Presence, contact: Optional["LegacyContact"] = None
    ) -> PepAvatar:
        if stanza.get_to() == self.xmpp.boundjid.bare:
            item = PepAvatar()
            if self.xmpp.avatar is not None:
                item.set_avatar_from_cache(self.xmpp.avatar)
            return item

        if contact is None:
            contact = await self.__get_contact(stanza)

        item = PepAvatar()
        if contact.stored.avatar is not None:
            stored = avatar_cache.get(contact.stored.avatar)
            assert stored is not None
            item.set_avatar_from_cache(stored)
        return item

    async def _get_authorized_nick(
        self, stanza: Iq | Presence, contact: Optional["LegacyContact"] = None
    ) -> UserNick:
        if stanza.get_to() == self.xmpp.boundjid.bare:
            return get_user_nick(self.xmpp.COMPONENT_NAME)

        if contact is None:
            contact = await self.__get_contact(stanza)

        if contact.name is not None:
            return get_user_nick(contact.name)
        else:
            return UserNick()

    def __reply_with(
        self, iq: Iq, content: AvatarData | AvatarMetadata | None, item_id: str | None
    ) -> None:
        requested_items = iq["pubsub"]["items"]

        if len(requested_items) == 0:
            self._reply_with_payload(iq, content, item_id)
        else:
            for item in requested_items:
                if item["id"] == item_id:
                    self._reply_with_payload(iq, content, item_id)
                    return
            else:
                raise XMPPError("item-not-found")

    async def _get_avatar_data(self, iq: Iq) -> None:
        pep_avatar = await self._get_authorized_avatar(iq)
        self.__reply_with(iq, pep_avatar.data, pep_avatar.id)

    async def _get_avatar_metadata(self, iq: Iq) -> None:
        pep_avatar = await self._get_authorized_avatar(iq)
        self.__reply_with(iq, pep_avatar.metadata, pep_avatar.id)

    async def _get_vcard(self, iq: Iq):
        # this is not the proper way that clients should retrieve VCards, but
        # gajim does it this way.
        # https://xmpp.org/extensions/xep-0292.html#sect-idm45669698174224
        session = self.xmpp.get_session_from_stanza(iq)
        contact = await session.contacts.by_jid(iq.get_to())
        vcard = await contact.get_vcard()
        if vcard is None:
            raise XMPPError("item-not-found")
        self._reply_with_payload(iq, vcard, "current", VCARD4_NAMESPACE)

    @staticmethod
    def _reply_with_payload(
        iq: Iq,
        payload: AvatarMetadata | AvatarData | VCard4 | None,
        id_: str | None,
        namespace: str | None = None,
    ) -> None:
        result = iq.reply()
        item = Item()
        if payload:
            item.set_payload(payload.xml)
            item["id"] = id_
            result["pubsub"]["items"]["node"] = (
                namespace if namespace else payload.namespace
            )
        result["pubsub"]["items"].append(item)
        result.send()

    async def __broadcast(
        self, data, from_: JidStr, to: OptJidStr = None, **kwargs
    ) -> None:
        from_ = JID(from_)
        if from_ != self.xmpp.boundjid.bare and to is not None:
            to = JID(to)
            session = self.xmpp.get_session_from_jid(to)
            if session is None:
                return
            await session.ready

        item = EventItem()
        if data:
            item.set_payload(data.xml)
        for k, v in kwargs.items():
            item[k] = v

        items = EventItems()
        items.append(item)
        items["node"] = kwargs.get("node") or data.namespace

        event = Event()
        event.append(items)

        msg = self.xmpp.Message()
        msg.set_type("headline")
        msg.set_from(from_)
        msg.append(event)

        if to is None:
            with self.xmpp.store.session() as orm:
                for u in orm.query(GatewayUser).all():
                    new_msg = copy(msg)
                    new_msg.set_to(u.jid.bare)
                    new_msg.send()
        else:
            msg.set_to(to)
            msg.send()

    async def broadcast_avatar(
        self, from_: JidStr, to: JidStr, cached_avatar: CachedAvatar | None
    ) -> None:
        if cached_avatar is None:
            await self.__broadcast(AvatarMetadata(), from_, to)
        else:
            pep_avatar = PepAvatar()
            pep_avatar.set_avatar_from_cache(cached_avatar)
            assert pep_avatar.metadata is not None
            await self.__broadcast(
                pep_avatar.metadata, from_, to, id=pep_avatar.metadata["info"]["id"]
            )

    def broadcast_nick(
        self,
        user_jid: JID,
        jid: JidStr,
        nick: str | None = None,
    ) -> None:
        jid = JID(jid)
        nickname = get_user_nick(nick)
        log.debug("New nickname: %s", nickname)
        self.xmpp.loop.create_task(self.__broadcast(nickname, jid, user_jid.bare))


def get_user_nick(nick: str | None = None) -> UserNick:
    user_nick = UserNick()
    if nick is not None:
        user_nick["nick"] = nick
    return user_nick


log = logging.getLogger(__name__)
register_plugin(PubSubComponent)



from abc import ABCMeta
from typing import TYPE_CHECKING

from slixmpp import JID

from ...util.types import MessageOrPresenceTypeVar

if TYPE_CHECKING:
    from ..gateway import BaseGateway
    from ..session import BaseSession


class MetaBase(ABCMeta):
    pass


class Base:
    session: "BaseSession" = NotImplemented
    xmpp: "BaseGateway" = NotImplemented

    jid: JID = NotImplemented
    name: str | None = NotImplemented

    @property
    def user_jid(self):
        return self.session.user_jid

    @property
    def user_pk(self):
        return self.session.user_pk


class BaseSender(Base):
    is_participant: bool = NotImplemented

    def _send(
        self, stanza: MessageOrPresenceTypeVar, **send_kwargs
    ) -> MessageOrPresenceTypeVar:
        raise NotImplementedError



import asyncio
import logging
import warnings
from collections.abc import AsyncIterator, Iterator
from typing import TYPE_CHECKING, Generic

from slixmpp import JID
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from sqlalchemy.orm import Session
from sqlalchemy.orm import Session as OrmSession

from ..db.models import Contact, GatewayUser
from ..util import SubclassableOnce
from ..util.jid_escaping import ESCAPE_TABLE, unescape_node
from ..util.lock import NamedLockMixin
from ..util.types import LegacyContactType, LegacyUserIdType
from ..util.util import timeit
from .contact import LegacyContact

if TYPE_CHECKING:
    from ..core.session import BaseSession


class ContactIsUser(Exception):
    pass


class LegacyRoster(
    Generic[LegacyUserIdType, LegacyContactType],
    NamedLockMixin,
    SubclassableOnce,
):
    """
    Virtual roster of a gateway user that allows to represent all
    of their contacts as singleton instances (if used properly and not too bugged).

    Every :class:`.BaseSession` instance will have its own :class:`.LegacyRoster` instance
    accessible via the :attr:`.BaseSession.contacts` attribute.

    Typically, you will mostly use the :meth:`.LegacyRoster.by_legacy_id` function to
    retrieve a contact instance.

    You might need to override :meth:`.LegacyRoster.legacy_id_to_jid_username` and/or
    :meth:`.LegacyRoster.jid_username_to_legacy_id` to incorporate some custom logic
    if you need some characters when translation JID user parts and legacy IDs.
    """

    _contact_cls: type[LegacyContactType]

    def __init__(self, session: "BaseSession") -> None:
        super().__init__()

        self.log = logging.getLogger(f"{session.user_jid.bare}:roster")
        self.user_legacy_id: LegacyUserIdType | None = None
        self.ready: asyncio.Future[bool] = session.xmpp.loop.create_future()

        self.session = session
        self.__filling = False

    @property
    def user(self) -> GatewayUser:
        return self.session.user

    def orm(self) -> Session:
        return self.session.xmpp.store.session()

    def from_store(self, stored: Contact) -> LegacyContactType:
        return self._contact_cls(self.session, stored=stored)

    def __repr__(self) -> str:
        return f"<Roster of {self.session.user_jid}>"

    def __iter__(self) -> Iterator[LegacyContactType]:
        with self.orm() as orm:
            for stored in orm.query(Contact).filter_by(user=self.user).all():
                if stored.updated:
                    yield self.from_store(stored)

    def known_contacts(self, only_friends: bool = True) -> dict[str, LegacyContactType]:
        if only_friends:
            return {c.jid.bare: c for c in self if c.is_friend}
        return {c.jid.bare: c for c in self}

    async def by_jid(self, contact_jid: JID) -> LegacyContactType:
        # """
        # Retrieve a contact by their JID
        #
        # If the contact was not instantiated before, it will be created
        # using :meth:`slidge.LegacyRoster.jid_username_to_legacy_id` to infer their
        # legacy user ID.
        #
        # :param contact_jid:
        # :return:
        # """
        username = contact_jid.node
        if not username:
            raise XMPPError(
                "bad-request", "Contacts must have a local part in their JID"
            )
        contact_jid = JID(contact_jid.bare)
        async with self.lock(("username", username)):
            legacy_id = await self.jid_username_to_legacy_id(username)
            if legacy_id == self.user_legacy_id:
                raise ContactIsUser
            if self.get_lock(("legacy_id", legacy_id)):
                self.log.debug("Already updating %s via by_legacy_id()", contact_jid)
                return await self.by_legacy_id(legacy_id)

            with self.orm() as orm:
                stored = (
                    orm.query(Contact)
                    .filter_by(user=self.user, jid=contact_jid)
                    .one_or_none()
                )
            if stored is None:
                stored = Contact(
                    user_account_id=self.session.user_pk,
                    legacy_id=legacy_id,
                    jid=contact_jid,
                )
            return await self.__update_if_needed(stored)

    async def __update_if_needed(self, stored: Contact) -> LegacyContactType:
        contact = self.from_store(stored)
        if contact.stored.updated:
            return contact

        with contact.updating_info(merge=False):
            await contact.update_info()
            if contact.is_friend and not self.__filling:
                await contact.add_to_roster()

        if contact.cached_presence is not None:
            contact._store_last_presence(contact.cached_presence)
        return contact

    def by_jid_only_if_exists(self, contact_jid: JID) -> LegacyContactType | None:
        with self.orm() as orm:
            stored = (
                orm.query(Contact)
                .filter_by(user=self.user, jid=contact_jid)
                .one_or_none()
            )
            if stored is not None and stored.updated:
                return self.from_store(stored)
        return None

    @timeit
    async def by_legacy_id(self, legacy_id: LegacyUserIdType) -> LegacyContactType:
        """
        Retrieve a contact by their legacy_id

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.legacy_id_to_jid_username` to infer their
        legacy user ID.

        :param legacy_id:
        :return:
        """
        if legacy_id == self.user_legacy_id:
            raise ContactIsUser
        async with self.lock(("legacy_id", legacy_id)):
            username = await self.legacy_id_to_jid_username(legacy_id)
            if self.get_lock(("username", username)):
                self.log.debug("Already updating %s via by_jid()", username)

                return await self.by_jid(
                    JID(username + "@" + self.session.xmpp.boundjid.bare)
                )

            with self.orm() as orm:
                stored = (
                    orm.query(Contact)
                    .filter_by(user=self.user, legacy_id=str(legacy_id))
                    .one_or_none()
                )
            if stored is None:
                stored = Contact(
                    user_account_id=self.session.user_pk,
                    legacy_id=str(legacy_id),
                    jid=JID(f"{username}@{self.session.xmpp.boundjid.bare}"),
                )
            return await self.__update_if_needed(stored)

    async def legacy_id_to_jid_username(self, legacy_id: LegacyUserIdType) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it is case-sensitive or contains
        forbidden characters not covered by :xep:`0106`.

        :param legacy_id:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, jid_username: str) -> LegacyUserIdType:
        """
        Convert a JID user part to a legacy ID.

        Should be overridden in case legacy IDs are not strings, or more generally
        for any case where the username part of a JID (unescaped with to the mapping
        defined by :xep:`0106`) is not enough to identify a contact on the legacy network.

        Default implementation is an identity operation

        :param jid_username: User part of a JID, ie "user" in "user@example.com"
        :return: An identifier for the user on the legacy network.
        """
        return unescape_node(jid_username)  # type:ignore

    @timeit
    async def _fill(self, orm: OrmSession):
        try:
            if hasattr(self.session.xmpp, "TEST_MODE"):
                # dirty hack to avoid mocking xmpp server replies to this
                # during tests
                raise PermissionError
            iq = await self.session.xmpp["xep_0356"].get_roster(
                self.session.user_jid.bare
            )
            user_roster = iq["roster"]["items"]
        except (PermissionError, IqError, IqTimeout):
            user_roster = None

        self.__filling = True
        async for contact in self.fill():
            if user_roster is None:
                continue
            item = contact.get_roster_item()
            old = user_roster.get(contact.jid.bare)
            if old is not None and all(
                old[k] == item[contact.jid.bare].get(k)
                for k in ("subscription", "groups", "name")
            ):
                self.log.debug("No need to update roster")
                continue
            self.log.debug("Updating roster")
            if not contact.is_friend:
                continue
            if not self.session.user.preferences.get("roster_push", True):
                continue
            try:
                await self.session.xmpp["xep_0356"].set_roster(
                    self.session.user_jid.bare,
                    item,
                )
            except (PermissionError, IqError, IqTimeout) as e:
                warnings.warn(f"Could not add to roster: {e}")
            else:
                contact.added_to_roster = True
                contact.send_last_presence(force=True)
        orm.commit()
        self.__filling = False

    async def fill(self) -> AsyncIterator[LegacyContact]:
        """
        Populate slidge's "virtual roster".

        This should yield contacts that are meant to be added to the user's
        roster, typically by using ``await self.by_legacy_id(contact_id)``.
        Setting the contact nicknames, avatar, etc. should be in
        :meth:`LegacyContact.update_info()`

        It's not mandatory to override this method, but it is recommended way
        to populate "friends" of the user. Calling
        ``await (await self.by_legacy_id(contact_id)).add_to_roster()``
        accomplishes the same thing, but doing it in here allows to batch
        DB queries and is better performance-wise.

        """
        return
        yield


log = logging.getLogger(__name__)



import unittest.mock

from slixmpp import JID
from slixmpp.plugins.xep_0356.permissions import IqPermission
from test_muc import Base as BaseMUC

from conftest import UUIDFixtureMixin

class MDSMixin(UUIDFixtureMixin):
    def setUp(self):
        super().setUp()
        for domain in "test.com", "montague.lit":
            self.xmpp["xep_0356"].granted_privileges[domain].iq[
                "http://jabber.org/protocol/pubsub"
            ] = IqPermission.BOTH
            self.xmpp["xep_0356"].granted_privileges[domain].iq[
                "http://jabber.org/protocol/pubsub#owner"
            ] = IqPermission.BOTH


class TestMDS(MDSMixin, BaseMUC):
    def test_add_to_whitelist(self):
        task = self.xmpp.loop.create_task(
            self.xmpp._BaseGateway__add_component_to_mds_whitelist(JID("test@test.com"))
        )
        self.send(  # language=XML
            """
            <iq id="1"
                to="test@test.com"
                from="aim.shakespeare.lit"
                type="set">
              <privileged_iq xmlns="urn:xmpp:privilege:2">
                <iq xmlns="jabber:client"
                    type="set"
                    to="test@test.com"
                    from="test@test.com"
                    id="1">
                  <pubsub xmlns="http://jabber.org/protocol/pubsub">
                    <create node="urn:xmpp:mds:displayed:0" />
                  </pubsub>
                </iq>
              </privileged_iq>
            </iq>
            """,
            use_values=False,
        )
        self.recv(  # language=XML
            """
            <iq id="1"
                type="result"
                to="test@test.com"
                from="aim.shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <iq xmlns="jabber:client"
                      id="1"
                      type="result"
                      to="test@localhost" />
                </forwarded>
              </privilege>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq id="2"
                to="test@test.com"
                from="aim.shakespeare.lit"
                type="set">
              <privileged_iq xmlns="urn:xmpp:privilege:2">
                <iq xmlns="jabber:client"
                    type="set"
                    to="test@test.com"
                    from="test@test.com"
                    id="2">
                  <pubsub xmlns="http://jabber.org/protocol/pubsub#owner">
                    <affiliations node="urn:xmpp:mds:displayed:0">
                      <affiliation jid="aim.shakespeare.lit"
                                   affiliation="member" />
                    </affiliations>
                  </pubsub>
                </iq>
              </privileged_iq>
            </iq>
            """,
            use_values=False,
        )
        self.recv(  # language=XML
            """
            <iq id="2"
                type="result"
                to="test@test.com"
                from="aim.shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <iq xmlns="jabber:client"
                      id="2"
                      type="result"
                      to="test@localhost" />
                </forwarded>
              </privilege>
            </iq>
            """
        )
        assert task.done()

    def test_receive_event(self):
        session = self.get_romeo_session()
        # juliet = self.juliet
        muc = self.get_private_muc()

        with unittest.mock.patch("test_muc.Session.on_displayed") as on_displayed:
            self.recv(  # language=XML
                f"""
            <message from='{session.user_jid}'
                     to='{self.xmpp.boundjid.bare}'
                     type='headline'
                     id='new-displayed-pep-event'>
              <event xmlns='http://jabber.org/protocol/pubsub#event'>
                <items node='urn:xmpp:mds:displayed:0'>
                  <item id='{muc.jid}'>
                    <displayed xmlns='urn:xmpp:mds:displayed:0'>
                      <stanza-id xmlns='urn:xmpp:sid:0'
                                 by='what@ev.er'
                                 id='1337' />
                    </displayed>
                  </item>
                </items>
              </event>
            </message>
            """
            )
            on_displayed.assert_awaited_once()
            assert on_displayed.call_args[0][0].jid == muc.jid
            assert on_displayed.call_args[0][1] == "legacy-1337"

    def test_send_mds(self):
        muc = self.get_private_muc()
        participant = self.run_coro(muc.get_user_participant())
        participant.displayed("legacy-msg-id")
        self.send(  # language=XML
            """
            <iq id="4"
                to="romeo@montague.lit"
                from="aim.shakespeare.lit"
                type="set">
              <privileged_iq xmlns="urn:xmpp:privilege:2">
                <iq xmlns="jabber:client"
                    to="romeo@montague.lit"
                    from="romeo@montague.lit"
                    id="4"
                    type="set">
                  <pubsub xmlns="http://jabber.org/protocol/pubsub">
                    <publish node="urn:xmpp:mds:displayed:0">
                      <item id="room-private@aim.shakespeare.lit">
                        <displayed xmlns="urn:xmpp:mds:displayed:0">
                          <stanza-id xmlns="urn:xmpp:sid:0"
                                     id="msg-id"
                                     by="room-private@aim.shakespeare.lit" />
                        </displayed>
                      </item>
                    </publish>
                    <publish-options>
                      <x xmlns="jabber:x:data"
                         type="submit">
                        <field var="FORM_TYPE"
                               type="hidden">
                          <value>http://jabber.org/protocol/pubsub#publish-options</value>
                        </field>
                        <field var="pubsub#persist_items">
                          <value>true</value>
                        </field>
                        <field var="pubsub#max_items">
                          <value>max</value>
                        </field>
                        <field var="pubsub#send_last_published_item">
                          <value>never</value>
                        </field>
                        <field var="pubsub#access_model">
                          <value>whitelist</value>
                        </field>
                      </x>
                    </publish-options>
                  </pubsub>
                </iq>
              </privileged_iq>
            </iq>
            """,
            use_values=False,
        )
        self.recv(  # language=XML
            """
            <iq id="4"
                from="romeo@montague.lit"
                to="aim.shakespeare.lit"
                type="result" />
            """
        )



from .chat_state import ChatStateMixin
from .marker import MarkerMixin
from .message import MessageContentMixin


class MessageMixin(ChatStateMixin, MarkerMixin, MessageContentMixin):
    __slots__: list[str] = []


__all__ = ("MessageMixin",)



import asyncio
import random
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from slixmpp.exceptions import XMPPError

from slidge.util.types import Hat, MessageReference

from .util import ASSETS_DIR, later

if TYPE_CHECKING:
    from .session import Session


@dataclass
class DirectMessage:
    sender: int
    text: str
    id: str
    reply_to: str | None = None


@dataclass
class GroupMessage:
    sender: int
    group_id: str
    text: str
    id: str
    reply_to: str | None = None


@dataclass
class Profile:
    nickname: str
    full_name: str
    avatar: Path
    avatar_unique_id: str


@dataclass
class GroupInfo:
    name: str
    avatar: Path | None = None
    avatar_unique_id: str | None = None


class SuperDuperClient:
    @staticmethod
    async def send_2fa(username, _password):
        if username != "slidger":
            raise XMPPError("not-authorized", "Use 'slidger' as the username or GTFO")

    @staticmethod
    async def validate_2fa(_username, _password, code: str):
        if code != "666":
            raise XMPPError("not-authorized", "Wrong code! It's 666.")

    def __init__(self, session: "Session"):
        self.session = session
        self.user_id = -1
        self.__task = asyncio.create_task(self.__incoming_messages())

    async def __incoming_messages(self):
        while True:
            await self.on_contact_message(
                DirectMessage(666, "Hey devil!", id=uuid.uuid4().hex)
            )
            await self.on_contact_message(
                DirectMessage(111, "Hey!", id=uuid.uuid4().hex)
            )
            await asyncio.sleep(300)
            await self.on_contact_message(
                DirectMessage(222, "Ho!", id=uuid.uuid4().hex)
            )
            await asyncio.sleep(300)

    async def login(self):
        pass

    async def send_direct_msg(self, text: str, contact_id: int):
        i = uuid.uuid4().hex
        later(
            self.on_contact_message(
                DirectMessage(
                    id=uuid.uuid4().hex, text="A reply", sender=contact_id, reply_to=i
                )
            )
        )
        return DirectMessage(text=text, id=i, sender=self.user_id)

    async def send_group_msg(self, text: str, group_id: str):
        i = uuid.uuid4().hex
        later(
            self.on_group_message(
                GroupMessage(
                    id=uuid.uuid4().hex,
                    text="A reply",
                    group_id=group_id,
                    sender=666,
                    reply_to=i,
                )
            )
        )
        return GroupMessage(text=text, id=i, sender=self.user_id, group_id=group_id)

    @staticmethod
    async def get_profile(user_id: int) -> Profile:
        return _PROFILES[user_id]

    @staticmethod
    async def get_group_info(group_id: str) -> GroupInfo:
        return _GROUPS[group_id]

    async def on_contact_message(self, msg: DirectMessage):
        contact = await self.session.contacts.by_legacy_id(msg.sender)
        contact.send_text(
            msg.text,
            msg.id,
            reply_to=MessageReference(msg.reply_to, author="user"),
        )

    async def on_group_message(self, msg: GroupMessage):
        muc = await self.session.bookmarks.by_legacy_id(msg.group_id)
        participant = await muc.get_participant_by_legacy_id(0)
        participant.send_text(
            msg.text,
            msg.id,
            reply_to=MessageReference(msg.reply_to, author="user"),
        )
        await asyncio.sleep(1)
        participant.send_text(
            msg.text + " (correction)",
            msg.id,
            reply_to=MessageReference(msg.reply_to, author="user"),
            correction=True,
        )
        user = await muc.get_user_participant()
        if random.random() < 0.5:
            user.set_hats([("prout", "prout")])
            participant.set_hats([])
        else:
            user.set_hats([])
            participant.set_hats(
                [
                    Hat("12", "gloup"),
                    Hat(
                        str(random.randint(0, 256)),
                        "gloup" + str(random.randint(0, 256)),
                    ),
                ]
            )


_PROFILES = {
    111: Profile(
        nickname="Baba",
        avatar=ASSETS_DIR / "slidge-mono-black.png",
        full_name="Ba Ba",
        avatar_unique_id="baba-uid",
    ),
    222: Profile(
        nickname="Bibi",
        avatar=ASSETS_DIR / "slidge-mono-white.png",
        full_name="Bi Bi",
        avatar_unique_id="bibi-uid",
    ),
    666: Profile(
        nickname="The devil",
        avatar=ASSETS_DIR / "5x5.png",
        full_name="Lucy Fer",
        avatar_unique_id="devil-uid",
    ),
    000: Profile(
        nickname="🎉 The joker 🎉",
        avatar=ASSETS_DIR / "5x5.png",
        full_name="A guy with emojis in his nick",
        avatar_unique_id="devil-uid",
    ),
}

_GROUPS = {
    "aaa": GroupInfo(
        name="The groupchat A",
        avatar=ASSETS_DIR / "slidge-color-small.png",
        avatar_unique_id="slidge-color-small",
    ),
    "bbb": GroupInfo(
        name="The groupchat B",
    ),
}



import unittest.mock
from contextlib import asynccontextmanager

import pytest
from conftest import AvatarFixtureMixin
from slixmpp import JID
from slixmpp import __version__ as slix_version
from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0060.stanza import EventItem
from slixmpp.plugins.xep_0084 import MetaData

from slidge import BaseGateway, BaseSession
from slidge.core.session import _sessions
from slidge.util.test import SlidgeTest
from slidge.util.types import LinkPreview, LegacyAttachment


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Session(BaseSession):
    async def login(self):
        return "YUP"


@pytest.mark.usefixtures("avatar")
class TestSession(AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        self.setup_logged_session()
        self.xmpp["xep_0060"].map_node_event(MetaData.namespace, "avatar_metadata")
        register_stanza_plugin(EventItem, MetaData)

    @staticmethod
    def get_romeo_session() -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    def test_gateway_receives_presence_probe(self):
        self.recv(  # language=XML
            f"""
            <presence from='romeo@montague.lit/dino'
                      to="{self.xmpp.boundjid.bare}"
                      type="probe" />
            """
        )
        self.send(  # language=XML
            f"""
            <presence to='romeo@montague.lit/dino'
                      from="{self.xmpp.boundjid.bare}">
              <status>YUP</status>
              <show>chat</show>
              <c xmlns="http://jabber.org/protocol/caps"
                 node="http://slixmpp.com/ver/{slix_version}"
                 hash="sha-1"
                 ver="AuL8MdHJviOT17Bh1mfkW7IM7NU=" />
            </presence>
            """
        )
        assert self.next_sent() is None

    def test_avatar(self):
        with unittest.mock.patch("slidge.BaseSession.on_avatar") as on_avatar:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     type="headline"
                     to="{self.xmpp.boundjid.bare}"
                     id="mid">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:avatar:metadata">
                  <item id="{self.avatar_sha1}"
                        publisher="test@localhost">
                    <metadata xmlns="urn:xmpp:avatar:metadata">
                      <info id="{self.avatar_sha1}"
                            height="5"
                            width="5"
                            type="image/png"
                            bytes="{len(self.avatar_bytes)}" />
                    </metadata>
                  </item>
                </items>
              </event>
            </message>
            """
            )
            self.send(  # language=XML
                f"""
            <iq type="get"
                to="romeo@montague.lit"
                from="{self.xmpp.boundjid}"
                id="2">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:data">
                  <item id="{self.avatar_sha1}" />
                </items>
              </pubsub>
            </iq>
            """
            )
            self.recv(  # language=XML
                f"""
            <iq type="result"
                from="romeo@montague.lit"
                id="2">
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='urn:xmpp:avatar:data'>
                  <item id='{self.avatar_sha1}'>
                    <data xmlns='urn:xmpp:avatar:data'>{self.avatar_base64}</data>
                  </item>
                </items>
              </pubsub>
            </iq>
            """
            )
            on_avatar.assert_awaited_with(
                self.avatar_bytes, self.avatar_sha1, "image/png", 5, 5
            )
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     type="headline"
                     to="{self.xmpp.boundjid.bare}"
                     id="mid">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:avatar:metadata">
                  <item id="{self.avatar_sha1}"
                        publisher="test@localhost">
                    <metadata xmlns="urn:xmpp:avatar:metadata">
                      <info id="{self.avatar_sha1}"
                            height="5"
                            width="5"
                            type="image/png"
                            bytes="{len(self.avatar_bytes)}" />
                    </metadata>
                  </item>
                </items>
              </event>
            </message>
            """
            )

    def test_avatar_unpublish(self):
        with unittest.mock.patch("slidge.BaseSession.on_avatar") as on_avatar:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     type="headline"
                     to="{self.xmpp.boundjid.bare}"
                     id="mid">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:avatar:metadata">
                  <item id="{self.avatar_sha1}"
                        publisher="test@localhost">
                    <metadata xmlns="urn:xmpp:avatar:metadata" />
                  </item>
                </items>
              </event>
            </message>
            """
            )
            on_avatar.assert_awaited_with(None, None, None, None, None)

    def test_user_send_invitation_to_standard_muc(self):
        self.recv(  # language=XML
            f"""
            <message from="romeo@montague.lit"
                     to="juliet@{self.xmpp.boundjid.bare}"
                     id="mid">
              <x xmlns='jabber:x:conference'
                 jid='darkcave@macbeth.shakespeare.lit'
                 password='cauldronburn'
                 reason='Hey Hecate, this is the place for all good witches!' />
            </message>
            """
        )
        msg = self.next_sent()
        assert msg["type"] == "error"
        assert msg["error"]["condition"] == "bad-request"

    def test_user_send_invitation(self):
        with unittest.mock.patch("slidge.BaseSession.on_invitation") as on_invitation:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     to="juliet@{self.xmpp.boundjid.bare}"
                     id="mid">
              <x xmlns='jabber:x:conference'
                 jid='room@{self.xmpp.boundjid.bare}'
                 reason='Hey Hecate, this is the place for all good witches!' />
            </message>
            """
            )
            on_invitation.assert_awaited_once()
            assert on_invitation.call_args[0][0].jid == self.juliet.jid
            assert on_invitation.call_args[0][1].jid == self.room.jid
            assert (
                on_invitation.call_args[0][2]
                == "Hey Hecate, this is the place for all good witches!"
            )

    def test_link_preview(self):
        with unittest.mock.patch("slidge.BaseSession.on_text") as on_text:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     to="juliet@{self.xmpp.boundjid.bare}"
                     id="mid">
              <body>I wanted to mention https://the.link.example.com/what-was-linked-to</body>
              <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                               xmlns:og="https://ogp.me/ns#"
                               rdf:about="https://the.link.example.com/what-was-linked-to">
                <og:title>Page Title</og:title>
                <og:description>Page Description</og:description>
                <og:url>Canonical URL</og:url>
                <og:image>https://link.to.example.com/image.png</og:image>
                <og:site_name>Some Website</og:site_name>
              </rdf:Description>
            </message>
            """
            )
            on_text.assert_awaited_once()
            args, kwargs = on_text.call_args
            assert args[0].jid == self.juliet.jid
            assert (
                args[1]
                == "I wanted to mention https://the.link.example.com/what-was-linked-to"
            )
            # kwargs = on_text.c
            assert kwargs == dict(
                reply_to_msg_id=None,
                reply_to_fallback_text=None,
                reply_to=None,
                thread=None,
                link_previews=[
                    LinkPreview(
                        about="https://the.link.example.com/what-was-linked-to",
                        title="Page Title",
                        description="Page Description",
                        url="Canonical URL",
                        image="https://link.to.example.com/image.png",
                        type=None,
                        site_name="Some Website",
                    )
                ],
            )

    def test_juliet_sends_link_preview(self):
        self.juliet.send_text(
            "I wanted to mention https://the.link.example.com/what-was-linked-to",
            link_previews=[
                LinkPreview(
                    about="https://the.link.example.com/what-was-linked-to",
                    title="Page Title",
                    description="Page Description",
                    url="Canonical URL",
                    image="https://link.to.example.com/image.png",
                    type=None,
                    site_name="Some Website",
                )
            ],
        )
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <body>I wanted to mention https://the.link.example.com/what-was-linked-to</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
              <Description xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                           about="https://the.link.example.com/what-was-linked-to">
                <title xmlns="https://ogp.me/ns#">Page Title</title>
                <description xmlns="https://ogp.me/ns#">Page Description</description>
                <url xmlns="https://ogp.me/ns#">Canonical URL</url>
                <image xmlns="https://ogp.me/ns#">https://link.to.example.com/image.png</image>
                <site_name xmlns="https://ogp.me/ns#">Some Website</site_name>
              </Description>
            </message>
            """
        )

    def test_mark_all_messages(self):
        self.xmpp.MARK_ALL_MESSAGES = True
        self.juliet.send_text("whatever", "msg_00")
        self.juliet.send_text("whatever", "msg_01")
        self.juliet.send_text("whatever", "msg_02")
        with unittest.mock.patch(
            "slidge.core.session.BaseSession.on_displayed"
        ) as on_displayed:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     to="{self.juliet.jid.bare}">
              <displayed xmlns='urn:xmpp:chat-markers:0'
                         id='msg_03' />
            </message>
            """
            )
        assert on_displayed.await_count == 3
        for i in range(3):
            assert on_displayed.call_args_list[i][0][1] == f"msg_0{i}"

    def test_mark_all_messages_duplicate(self):
        self.xmpp.MARK_ALL_MESSAGES = True
        self.juliet.send_text("whatever", "msg_00")
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="msg_00"
                     to="romeo@montague.lit">
              <body>whatever</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
            </message>
            """,
            use_values=False,
        )
        self.juliet.send_text("whatever", "msg_00")
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="msg_00"
                     to="romeo@montague.lit">
              <body>whatever</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
            </message>
            """,
            use_values=False,
        )


    def test_movim_sticker(self):
        sticker_stanza = f"""
            <message from="romeo@montague.lit/movim"
                     to="{self.juliet.jid.bare}">
              <body>Un autocollant a été envoyé via Movim</body>
              <html xmlns="http://jabber.org/protocol/xhtml-im">
                <body xmlns="http://www.w3.org/1999/xhtml">
                  <p>
                    <img src="cid:sha1+4b97ce7f0f06a0e05999f3c719cd5b4f3da992a7@bob.xmpp.org"
                         alt="Sticker" />
                  </p>
                </body>
              </html>
            </message>
            """
        self.recv(sticker_stanza)
        self.send(  # language=XML
            f"""
            <iq id="2"
                type="get"
                to="romeo@montague.lit/movim"
                from="{self.xmpp.boundjid.bare}">
              <data xmlns="urn:xmpp:bob"
                    cid="sha1+4b97ce7f0f06a0e05999f3c719cd5b4f3da992a7@bob.xmpp.org" />
            </iq>
            """
        )
        with unittest.mock.patch(
            "slidge.core.session.BaseSession.on_sticker"
        ) as on_sticker:
            self.recv(  # language=XML
                f"""
            <iq from='romeo@montague.lit/movim'
                id='2'
                to='{self.xmpp.boundjid.bare}'
                type='result'>
              <data xmlns='urn:xmpp:bob'
                    cid='sha-1+4b97ce7f0f06a0e05999f3c719cd5b4f3da992a7@bob.xmpp.org'
                    max-age='86400'
                    type='image/png'>iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABGdBTUEAALGPC/xhBQAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB9YGARc5KB0XV+IAAAAddEVYdENvbW1lbnQAQ3JlYXRlZCB3aXRoIFRoZSBHSU1Q72QlbgAAAF1JREFUGNO9zL0NglAAxPEfdLTs4BZM4DIO4C7OwQg2JoQ9LE1exdlYvBBeZ7jqch9//q1uH4TLzw4d6+ErXMMcXuHWxId3KOETnnXXV6MJpcq2MLaI97CER3N0vr4MkhoXe0rZigAAAABJRU5ErkJggg==</data>
            </iq>
            """
            )
            on_sticker.assert_awaited_once()
            args, kwargs = on_sticker.call_args
            chat, sticker = args
            assert chat.legacy_id == self.juliet.legacy_id
            assert sticker.hashes["sha_1"] == "4b97ce7f0f06a0e05999f3c719cd5b4f3da992a7"
            assert sticker.path.exists()
            assert sticker.content_type == "image/png"
        # this time slidge must have cached the BoBd ata, so no bob-fetching IQ
        with unittest.mock.patch(
            "slidge.core.session.BaseSession.on_sticker"
        ) as on_sticker:
            self.recv(sticker_stanza)
            on_sticker.assert_awaited_once()
            args, kwargs = on_sticker.call_args
            chat, sticker = args
            assert chat.legacy_id == self.juliet.legacy_id
            assert sticker.hashes["sha_1"] == "4b97ce7f0f06a0e05999f3c719cd5b4f3da992a7"
            assert sticker.path.exists()
        assert self.next_sent() is None

    def test_movim_custom_emoji(self):
        with unittest.mock.patch("slidge.core.session.BaseSession.on_text") as on_text:
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit/movim"
                     to="{self.juliet.jid.bare}">
              <body>fdsf :amogus:</body>
              <html xmlns="http://jabber.org/protocol/xhtml-im">
                <body xmlns="http://www.w3.org/1999/xhtml">
                  <p>fdsf
                  <img src="cid:sha-256+583ca9a99f6cd8454c24d81a43d913a98dd80f282ce5c8f0f8ede418990134af@bob.xmpp.org"
                       alt=":amogus:" /></p>
                </body>
              </html>
            </message>
            """
            )
        on_text.assert_awaited_once()
        args, kwargs = on_text.call_args
        assert args[1] == "fdsf :amogus:"

    def test_bob_not_found(self):
        self.recv(  # language=XML
            f"""
            <iq from='{self.romeo.user_jid}/movim'
                id='get-data-1'
                to='juliet@{self.xmpp.boundjid.bare}'
                type='get'>
              <data xmlns='urn:xmpp:bob'
                    cid='bogus' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="juliet@aim.shakespeare.lit"
                id="get-data-1"
                to="romeo@montague.lit/movim"
                type="error">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Bits of binary 'bogus' is not available</text>
              </error>
            </iq>
            """
        )

    def test_carbon_retract(self):
        with (
            unittest.mock.patch(
                "slidge.core.session.BaseSession.on_retract"
            ) as on_retract,
            unittest.mock.patch(
                "slidge.core.session.BaseSession.on_correct"
            ) as on_correct,
        ):
            self.juliet.retract("some-id", carbon=True)
            self.recv(  # language=XML
                f"""
            <message type="chat"
                     to="{self.juliet.jid.bare}"
                     from="romeo@montague.lit/movim"
                     id="slidge-carbon-whatever">
              <body>/me retracted the message 1269564719166132224</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:message-retract:1" />
              <retract xmlns="urn:xmpp:message-retract:1"
                       id="some-id" />
              <replace xmlns="urn:xmpp:message-correct:0"
                       id="some-id" />
            </message>
            """
            )
            on_correct.assert_not_awaited()
            on_retract.assert_not_awaited()

    def test_new_thread_from_xmpp(self):
        with (
            unittest.mock.patch("slidge.core.session.BaseSession.on_text") as on_text,
            unittest.mock.patch(
                "slidge.contact.contact.LegacyContact.create_thread",
                return_value="legacy-thread-id",
            ),
        ):
            self.recv(  # language=XML
                f"""
            <message type="chat"
                     to="{self.juliet.jid.bare}"
                     from="romeo@montague.lit/movim"
                     id="xmpp-msg-id">
              <body>I start a new thread</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <thread>xmpp-thread-id</thread>
            </message>
            """
            )
            on_text.assert_awaited_once()
            args, kwargs = on_text.call_args
            assert kwargs["thread"] == "legacy-thread-id"
        with unittest.mock.patch("slidge.core.session.BaseSession.on_text") as on_text:
            self.recv(  # language=XML
                f"""
            <message type="chat"
                     to="{self.juliet.jid.bare}"
                     from="romeo@montague.lit/movim"
                     id="xmpp-msg-id-2">
              <body>I send a new message in the new thread</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <thread>xmpp-thread-id</thread>
            </message>
            """
            )
            on_text.assert_awaited_once()
            args, kwargs = on_text.call_args
            assert kwargs["thread"] == "legacy-thread-id"

    def test_multi_correction_caption(self):
        from slidge import global_config

        global_config.USE_ATTACHMENT_ORIGINAL_URLS = True
        self.xmpp.use_message_ids = True

        self.run_coro(
            self.juliet.send_file(
                LegacyAttachment(url=self.avatar_url, caption="prout"),
                legacy_msg_id="original-id",
            )
        )
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit"
                     id="original-id">
              <x xmlns="jabber:x:oob">
                <url>AVATAR_URL</url>
              </x>
              <body>AVATAR_URL</body>
            </message>
            """,
            use_values=False,
        )
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="3"
                     to="romeo@montague.lit">
              <body>prout</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """,
            use_values=False,
        )

        self.run_coro(
            self.juliet.send_file(
                LegacyAttachment(
                    url=self.avatar_url + "--NEW",
                    caption="prout",
                ),
                legacy_msg_id="original-id",
                correction=True,
            )
        )

        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="4"
                     to="romeo@montague.lit">
              <body>/me retracted the message 3</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:message-retract:1" />
              <retract xmlns="urn:xmpp:message-retract:1"
                       id="3" />
              <replace xmlns="urn:xmpp:message-correct:0"
                       id="3" />
            </message>
            """,
            use_values=False,
        )

        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit"
                     id="5">
              <x xmlns="jabber:x:oob">
                <url>AVATAR_URL--NEW</url>
              </x>
              <body>AVATAR_URL--NEW</body>
              <replace xmlns="urn:xmpp:message-correct:0"
                       id="original-id" />
            </message>
            """,
            use_values=False,
        )
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="6"
                     to="romeo@montague.lit">
              <body>prout</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """,
            use_values=False,
        )

        self.xmpp.use_message_ids = False
        global_config.USE_ATTACHMENT_ORIGINAL_URLS = False

    def test_multi_correction(self):
        from slidge import global_config

        global_config.USE_ATTACHMENT_ORIGINAL_URLS = True
        self.xmpp.use_message_ids = True

        self.run_coro(
            self.juliet.send_file(self.avatar_url, legacy_msg_id="original-id")
        )
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit"
                     id="original-id">
              <x xmlns="jabber:x:oob">
                <url>AVATAR_URL</url>
              </x>
              <body>AVATAR_URL</body>
            </message>
            """,
            use_values=False,
        )

        self.run_coro(
            self.juliet.send_file(
                self.avatar_url + "--NEW",
                legacy_msg_id="original-id",
                correction=True,
            )
        )

        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     id="3"
                     to="romeo@montague.lit">
              <x xmlns="jabber:x:oob">
                <url>AVATAR_URL--NEW</url>
              </x>
              <body>AVATAR_URL--NEW</body>
              <replace xmlns="urn:xmpp:message-correct:0"
                       id="original-id" />
            </message>
            """,
            use_values=False,
        )

        self.xmpp.use_message_ids = False
        global_config.USE_ATTACHMENT_ORIGINAL_URLS = False

    def test_sims(self):
        class MockResponse:
            status = 200

        def mock_get():
            @asynccontextmanager
            async def mock_response(self, url):
                yield MockResponse

            return mock_response

        with (
            unittest.mock.patch("slidge.BaseSession.on_file") as on_file,
            unittest.mock.patch("aiohttp.ClientSession.get", new_callable=mock_get),
        ):
            self.recv(  # language=XML
                f"""
            <message from="romeo@montague.lit"
                     to="juliet@{self.xmpp.boundjid.bare}"
                     id="mid">
              <body>Look at the nice view from the summit.</body>
              <reference xmlns='urn:xmpp:reference:0'
                         begin='17'
                         end='20'
                         type='data'>
                <media-sharing xmlns='urn:xmpp:sims:1'>
                  <file xmlns='urn:xmpp:jingle:apps:file-transfer:5'>
                    <media-type>image/jpeg</media-type>
                    <name>summit.jpg</name>
                    <size>3032449</size>
                    <hash xmlns='urn:xmpp:hashes:2'
                          algo='sha3-256'>2XarmwTlNxDAMkvymloX3S5+VbylNrJt/l5QyPa+YoU=</hash>
                    <hash xmlns='urn:xmpp:hashes:2'
                          algo='id-blake2b256'>2AfMGH8O7UNPTvUVAM9aK13mpCY=</hash>
                    <desc>Photo from the summit.</desc>
                    <thumbnail xmlns='urn:xmpp:thumbs:1'
                               uri='cid:sha1+ffd7c8d28e9c5e82afea41f97108c6b4@bob.xmpp.org'
                               media-type='image/png'
                               width='128'
                               height='96' />
                  </file>
                  <sources>
                    <reference xmlns='urn:xmpp:reference:0'
                               type='data'
                               uri='https://download.montague.lit/4a771ac1-f0b2-4a4a-9700-f2a26fa2bb67/summit.jpg' />
                    <reference xmlns='urn:xmpp:reference:0'
                               type='data'
                               uri='xmpp:romeo@montague.lit/resource?jingle;id=9559976B-3FBF-4E7E-B457-2DAA225972BB' />
                  </sources>
                </media-sharing>
              </reference>
            </message>
            """
            )
            on_file.assert_awaited_once()
            assert (
                on_file.call_args.args[1]
                == "https://download.montague.lit/4a771ac1-f0b2-4a4a-9700-f2a26fa2bb67/summit.jpg"
            )

    def test_user_retract(self):
        with unittest.mock.patch("slidge.core.session.BaseSession.on_text") as on_text:
            on_text.return_value = "legacy-id"
            self.recv(  # language=XML
                """
            <message type="chat"
                     to="juliet@aim.shakespeare.lit"
                     from="romeo@montague.lit/whatevs"
                     id="origin-id">
              <body>hoy</body>
            </message>
            """
            )
            on_text.assert_awaited_once()
            contact = on_text.call_args[0][0]
            text = on_text.call_args[0][1]
            assert contact.jid.node == "juliet"
            assert text == "hoy"
        with unittest.mock.patch("slidge.core.session.BaseSession.on_retract") as on_retract:
            self.recv(  # language=XML
                """
            <message type="chat"
                     to="juliet@aim.shakespeare.lit"
                     from="romeo@montague.lit/whatevs">
              <retract id="origin-id"
                       xmlns='urn:xmpp:message-retract:1' />
            </message>
            """
            )
            on_text.assert_awaited_once()
            contact = on_retract.call_args[0][0]
            legacy_msg_id = on_retract.call_args[0][1]
            assert contact.jid.node == "juliet"
            assert legacy_msg_id == "legacy-id"

    def test_reject_omemo(self):
        elements = [
            "<encrypted />",
            "<encrypted xmlns='whatever' />",
            "<encrypted>blabla</encrypted>",
            "<encrypted xmlns='whatever'>blabla</encrypted>",
        ]
        for el in elements:
            with unittest.mock.patch("slidge.core.session.BaseSession.on_text") as on_text:
                self.recv(  # language=XML
                    f"""
            <message type="chat"
                     to="juliet@aim.shakespeare.lit"
                     from="romeo@montague.lit/whatevs"
                     id="origin-id">{el}
            <body>your client blablabla</body></message>
            """
                )
                on_text.assert_not_awaited()
                self.send(  # language=XML
                    """
            <message xmlns="jabber:component:accept"
                     type="error"
                     to="romeo@montague.lit/whatevs"
                     from="juliet@aim.shakespeare.lit"
                     id="origin-id">
              <error type="modify">
                <bad-request xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">You cannot send encrypted messages through this gateway</text>
              </error>
            </message>
            """,
                    use_values=False,
                )



import unittest.mock

import pytest
from conftest import AvatarFixtureMixin
from test_shakespeare import Base as BaseNoMUC

from slidge import LegacyMUC, MucType
from slidge.util.types import Avatar


@pytest.mark.usefixtures("avatar")
class TestContactAvatar(BaseNoMUC, AvatarFixtureMixin):
    def setUp(self):
        super().setUp()
        self.juliet.is_friend = True
        self.juliet.added_to_roster = True

    def __assert_not_found(self):
        juliet = self.juliet
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='{juliet.user_jid}/client'
                to='{juliet.jid.bare}'
                id='retrieve1'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='urn:xmpp:avatar:data'>
                  <item id='{self.avatar_sha1}' />
                </items>
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq xmlns="jabber:component:accept"
                type="error"
                from="juliet@aim.shakespeare.lit"
                to="{juliet.user_jid}/client"
                id="retrieve1">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
              </error>
            </iq>
            """
        )

    def __assert_publish(self, rewritten=False):
        h = self.avatar_sha1 if rewritten else self.avatar_original_sha1
        length = (
            len(self.avatar_bytes) if rewritten else len(self.avatar_path.read_bytes())
        )
        self.send(  # language=XML
            f"""
            <message type="headline"
                     from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:avatar:metadata">
                  <item id="{h}">
                    <metadata xmlns="urn:xmpp:avatar:metadata">
                      <info id="{h}"
                            type="image/png"
                            bytes="{length}"
                            height="5"
                            width="5" />
                    </metadata>
                  </item>
                </items>
              </event>
            </message>
            """
        )
        assert self.next_sent() is None

    def __assert_publish_empty(self):
        self.send(  # language=XML
            f"""
            <message type="headline"
                     from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:avatar:metadata">
                  <item>
                    <metadata xmlns="urn:xmpp:avatar:metadata" />
                  </item>
                </items>
              </event>
            </message>
            """,
            use_values=False,
        )
        assert self.next_sent() is None

    def test_avatar_path_no_id(self):
        juliet = self.juliet
        assert juliet.avatar is None

        juliet.avatar = None
        self.run_coro(juliet._set_avatar_task)
        assert self.next_sent() is None

        self.__assert_not_found()

        juliet.avatar = Avatar(path=self.avatar_path)
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish()

        juliet.avatar = Avatar(path=self.avatar_path)
        self.run_coro(juliet._set_avatar_task)
        assert self.next_sent() is None

        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path)))
        assert self.next_sent() is None

        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path)))
        assert self.next_sent() is None

        juliet.avatar = Avatar(path=self.avatar_path)
        self.run_coro(juliet._set_avatar_task)
        assert self.next_sent() is None

        juliet.avatar = None
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish_empty()

        self.run_coro(juliet.set_avatar(None))
        assert self.next_sent() is None

        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path)))
        self.__assert_publish()

        juliet.avatar = None
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish_empty()

    def test_avatar_path_with_id(self):
        juliet = self.juliet
        assert juliet.avatar is None

        self.xmpp.AVATAR_ID_TYPE = int
        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path, unique_id=123)))
        self.__assert_publish(rewritten=False)

        assert juliet.avatar.unique_id == 123

        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path, unique_id=123)))
        assert self.next_sent() is None

        assert juliet.avatar.unique_id == 123

        self.xmpp.AVATAR_ID_TYPE = str
        self.run_coro(juliet.set_avatar(Avatar(path=self.avatar_path, unique_id="123")))
        assert self.next_sent() is None
        assert juliet.avatar.unique_id == "123"

        self.run_coro(juliet.set_avatar(None))
        self.__assert_publish_empty()

        assert juliet.avatar is None

    def test_same_avatar_with_different_legacy_ids(self):
        self.run_coro(self.juliet.set_avatar(Avatar(path=self.avatar_path, unique_id="123")))
        self.__assert_publish(rewritten=False)
        assert self.juliet.avatar.unique_id == "123"
        assert self.juliet.get_avatar().id == self.avatar_original_sha1

        self.run_coro(self.juliet.set_avatar(Avatar(path=self.avatar_path, unique_id="456")))
        self.__assert_publish(
            rewritten=False
        )  # FIXME: ideally, we should not publish here…
        assert self.juliet.avatar.unique_id == "456"

        assert self.juliet.get_avatar().id == self.avatar_original_sha1

    def test_avatar_with_url(self):
        juliet = self.juliet
        assert juliet.avatar is None
        juliet.avatar = Avatar(url=self.avatar_url)
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish(rewritten=True)

        juliet.avatar = Avatar(url=self.avatar_url)
        self.run_coro(juliet._set_avatar_task)
        assert self.next_sent() is None

    def test_avatar_with_url_and_unique_id(self):
        juliet = self.juliet
        assert juliet.avatar is None
        self.run_coro(juliet.set_avatar(Avatar(url=self.avatar_url, unique_id="someid")))
        self.__assert_publish(rewritten=True)
        self.run_coro(juliet.set_avatar(Avatar(url=self.avatar_url, unique_id="someid")))
        assert self.next_sent() is None

    def test_url_no_headers_remote_has_not_changed(self):
        juliet = self.juliet
        juliet.avatar = Avatar(url=self.avatar_url_no_headers)
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish(rewritten=True)

        juliet = self.juliet
        juliet.avatar = Avatar(url=self.avatar_url_no_headers)
        self.run_coro(juliet._set_avatar_task)
        # TODO: no reason to broadcast the avatar here, but we rely on HTTP headers
        #       to not re-publish. This is probably in the realm of micro-optimisation
        #       and does not matter much though…
        self.__assert_publish(rewritten=True)

    def test_url_no_headers_remote_has_changed(self):
        juliet = self.juliet
        juliet.avatar = Avatar(url=self.avatar_url_no_headers)
        self.run_coro(juliet._set_avatar_task)
        self.__assert_publish(rewritten=True)

        with self.switch_avatar():
            juliet = self.juliet
            juliet.avatar = Avatar(url=self.avatar_url_no_headers)
            self.run_coro(juliet._set_avatar_task)
            assert self.next_sent() is not None

            juliet = self.juliet
            juliet.avatar = Avatar(url=self.avatar_url_no_headers)
            self.run_coro(juliet._set_avatar_task)
            assert self.next_sent() is not None

class MUC(LegacyMUC):
    type = MucType.GROUP
    user_nick = "romeo"

    async def update_info(self):
        return

class BaseMUC(BaseNoMUC):
    plugin = BaseNoMUC.plugin | {"MUC": MUC}

    def setUp(self):
        self.patch = unittest.mock.patch(
            "uuid.uuid4", return_value="uuid4"
        )
        self.patch.start()
        super().setUp()

    def tearDown(self):
        super().tearDown()
        self.patch.stop()

    def _assert_send_room_avatar(self, empty=False, url=False):
        if empty:
            photo = "<photo />"
        else:
            photo = f"<photo>{self.avatar_sha1 if url else self.avatar_original_sha1}</photo>"
        self.send(  # language=XML
            f"""
            <presence to="romeo@montague.lit/gajim"
                      from="room@aim.shakespeare.lit">
              <x xmlns="vcard-temp:x:update">{photo}</x>
            </presence>
            """,
            use_values=not empty,
        )

    def romeo_joins(self, muc: MUC):
        session = self.get_romeo_session()
        self.recv(  # language=XML
            f"""
            <presence from="{session.user_jid}/gajim"
                      to="room@{session.xmpp.boundjid.bare}/romeo">
              <x xmlns='http://jabber.org/protocol/muc' />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room@aim.shakespeare.lit/romeo"
                      to="romeo@montague.lit/gajim">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant"
                      jid="romeo@montague.lit/gajim" />
                <status code="100" />
                <status code="110" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </presence>
            """
        )
        subject_msg = self.next_sent()
        assert subject_msg.xml.find(f"{{{subject_msg.namespace}}}subject") is not None
        # assert self.next_sent()["from"] == "room@aim.shakespeare.lit"

    def get_muc(self, joined=True) -> MUC:
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        if joined:
            self.romeo_joins(muc)
        return self.run_coro(session.bookmarks.by_legacy_id("room"))


@pytest.mark.usefixtures("avatar")
class TestParticipantAvatar(BaseMUC, AvatarFixtureMixin):
    def romeo_joins(self, muc: MUC):
        super().romeo_joins(muc)
        self._assert_send_room_avatar(empty=True)

    def _assert_juliet_presence_no_avatar(self):
        self.send(  # language=XML
            """
            <presence from="room@aim.shakespeare.lit/juliet"
                      to="romeo@montague.lit/gajim">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant"
                      jid="juliet@aim.shakespeare.lit/slidge" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="juliet@aim.shakespeare.lit/slidge" />
            </presence>
            """
        )

    def _assert_juliet_presence_avatar(self, sha=None, url=False):
        self.send(  # language=XML
            f"""
            <presence from="room@aim.shakespeare.lit/juliet"
                      to="romeo@montague.lit/gajim">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant"
                      jid="juliet@aim.shakespeare.lit/slidge" />
              </x>
              <x xmlns="vcard-temp:x:update">
                <photo>{self.avatar_sha1 if url else self.avatar_original_sha1}</photo>
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="juliet@aim.shakespeare.lit/slidge" />
            </presence>
            """
        )

    def test_romeo_join_empty_room_then_juliet_joins_then_set_avatar(self):
        self.get_muc(joined=True)
        session = self.get_romeo_session()

        session.contacts.ready.set_result(True)
        muc = self.get_muc(joined=False)
        self.run_coro(muc.get_participant_by_contact(self.juliet))
        self._assert_juliet_presence_no_avatar()
        assert self.next_sent() is None

        juliet = self.juliet
        juliet.avatar = Avatar(path=self.avatar_path)
        # no broadcast of the contact avatar because not added to roster,
        # only the participant
        self.run_coro(juliet._set_avatar_task)
        self._assert_juliet_presence_avatar()
        assert self.next_sent() is None

        self.juliet.avatar = Avatar(path=self.avatar_path)
        assert self.next_sent() is None

        juliet = self.juliet
        juliet.avatar = None
        self.run_coro(juliet._set_avatar_task)
        self._assert_juliet_presence_no_avatar()
        assert self.next_sent() is None

    def test_romeo_join_empty_room_then_juliet_joins_then_set_avatar_with_url(self):
        self.get_muc(joined=True)
        session = self.get_romeo_session()

        session.contacts.ready.set_result(True)
        juliet = self.juliet
        self.run_coro(self.get_muc(joined=False).get_participant_by_contact(juliet))
        self._assert_juliet_presence_no_avatar()
        assert self.next_sent() is None
        juliet = self.juliet
        juliet.avatar = Avatar(url=self.avatar_url)
        # no broadcast of the contact avatar because not added to roster,
        # only the participant
        self.run_coro(juliet._set_avatar_task)
        self._assert_juliet_presence_avatar(url=True)
        assert self.next_sent() is None

        juliet.avatar = Avatar(url=self.avatar_url)
        self.run_coro(juliet._set_avatar_task)
        assert self.next_sent() is None

        juliet.avatar = None
        self.run_coro(juliet._set_avatar_task)
        self._assert_juliet_presence_no_avatar()
        assert self.next_sent() is None

    def test_avatar_forbidden_emoji_in_participant_nickname(self):
        self.get_muc(joined=True)
        session = self.get_romeo_session()
        juliet = self.juliet
        juliet.name = "juliet🎉"
        juliet.avatar = Avatar(url=self.avatar_url)
        session.contacts.ready.set_result(True)
        self.run_coro(self.get_muc(False).get_participant_by_contact(juliet))
        self.send(  # language=XML
            f"""
            <presence from="room@aim.shakespeare.lit/juliet-1934e"
                      to="romeo@montague.lit/gajim">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant"
                      jid="juliet@aim.shakespeare.lit/slidge" />
              </x>
              <x xmlns="vcard-temp:x:update">
                <photo>{self.avatar_sha1}</photo>
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="juliet@aim.shakespeare.lit/slidge" />
              <nick xmlns="http://jabber.org/protocol/nick">juliet🎉</nick>
            </presence>
            """
        )
        self.recv(  # language=XML
            """
            <iq from="romeo@montague.lit/gajim"
                to="room@aim.shakespeare.lit/juliet-1934e"
                type="get">
              <vCard xmlns="vcard-temp" />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq from="room@aim.shakespeare.lit/juliet-1934e"
                to="romeo@montague.lit/gajim"
                type="result"
                id="1">
              <vCard xmlns="vcard-temp">
                <PHOTO>
                  <BINVAL>{self.avatar_base64}</BINVAL>
                  <TYPE>image/png</TYPE>
                </PHOTO>
              </vCard>
            </iq>
            """
        )
        assert self.next_sent() is None
        juliet = self.run_coro(self.get_muc(False).get_participant_by_contact(juliet))
        juliet.send_text("prout")
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="groupchat"
                     from="room@aim.shakespeare.lit/juliet-1934e"
                     to="romeo@montague.lit/gajim">
              <body>prout</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid4"
                         by="room@aim.shakespeare.lit" />
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="juliet@aim.shakespeare.lit/slidge" />
              <nick xmlns="http://jabber.org/protocol/nick">juliet🎉</nick>
            </message>
            """
        )


@pytest.mark.usefixtures("avatar")
class TestRoomAvatar(BaseMUC, AvatarFixtureMixin):
    def test_room_avatar_change_after_join(self):
        self.get_muc(joined=True)
        self._assert_send_room_avatar(empty=True)
        muc = self.get_muc(joined=False)
        muc.avatar = self.avatar_path
        self.run_coro(muc._set_avatar_task)
        self.send(  # language=XML
            """
            <message type="groupchat"
                     from="room@aim.shakespeare.lit"
                     to="romeo@montague.lit/gajim">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid4"
                         by="room@aim.shakespeare.lit" />
              <x xmlns="http://jabber.org/protocol/muc#user">
                <status code="104" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
            </message>
            """,
            use_values=False,
        )
        self._assert_send_room_avatar()

    def test_room_avatar_on_join(self):
        muc = self.get_muc(joined=False)
        muc.avatar = self.avatar_path
        self.run_coro(muc._set_avatar_task)
        self.romeo_joins(muc)
        self._assert_send_room_avatar()

    def test_room_avatar_with_url(self):
        muc = self.get_muc(joined=False)
        muc.avatar = self.avatar_url
        self.run_coro(muc._set_avatar_task)
        self.romeo_joins(muc)
        self._assert_send_room_avatar(url=True)

    def test_room_avatar_with_url_and_unique_id(self):
        muc = self.get_muc(joined=False)
        self.run_coro(muc.set_avatar(Avatar(url=self.avatar_url,unique_id= "id")))
        self.romeo_joins(muc)
        self._assert_send_room_avatar(url=True)
        self.run_coro(muc.set_avatar(Avatar(url=self.avatar_url,unique_id= "id")))
        assert self.next_sent() is None



import asyncio
import hashlib
import io
import unittest
from base64 import b64encode
from contextlib import asynccontextmanager, contextmanager
from http import HTTPStatus
from pathlib import Path
from typing import Callable
from unittest.mock import patch

import pytest
from PIL import Image

from slidge.util import SubclassableOnce

SubclassableOnce.TEST_MODE = True


@pytest.fixture
def MockRE():
    class MockRE:
        @staticmethod
        def match(*a, **kw):
            return True

    return MockRE


@pytest.fixture(scope="session")
def avatar_path() -> Path:
    return Path(__file__).parent.parent / "dev" / "assets" / "5x5.png"


@pytest.fixture(scope="session")
def avatar2_path() -> Path:
    return Path(__file__).parent.parent / "dev" / "assets" / "slidge-color-small.png"


@pytest.fixture(scope="class")
def avatar(request, avatar_path, avatar2_path):
    img = Image.open(avatar_path)
    with io.BytesIO() as f:
        img.save(f, format="PNG")
        img_bytes = f.getvalue()

    class MockResponse:
        def __init__(self, status, headers=True):
            if headers:
                self.headers = {"etag": "etag", "last-modified": "last"}
            else:
                self.headers = {}
            self.status = status

        @staticmethod
        async def read():
            return img_bytes

        def raise_for_status(self):
            pass

    @asynccontextmanager
    async def mock_get(url, headers=None):
        if url == "SLOW":
            await asyncio.sleep(1)
        else:
            assert url in ("AVATAR_URL", "AVATAR_URL_NO_HEADERS")
        if headers and (
            headers.get("If-None-Match") == "etag"
            or headers.get("If-Modified-Since") == "last"
        ):
            yield MockResponse(HTTPStatus.NOT_MODIFIED)
        else:
            yield MockResponse(HTTPStatus.OK, headers=url == "AVATAR_URL")

    @contextmanager
    def switch_avatar(_self):
        nonlocal img_bytes
        backup = img_bytes
        with io.BytesIO() as f:
            Image.open(avatar2_path).save(f, format="PNG")
            img_bytes = f.getvalue()
        yield
        img_bytes = backup

    request.cls.avatar_path = avatar_path
    request.cls.avatar_image = img
    request.cls.avatar_bytes = img_bytes
    request.cls.avatar_sha1 = hashlib.sha1(img_bytes).hexdigest()
    request.cls.avatar_url = "AVATAR_URL"
    request.cls.avatar_url_no_headers = "AVATAR_URL_NO_HEADERS"
    request.cls.switch_avatar = switch_avatar

    request.cls.avatar_base64 = b64encode(img_bytes).decode("utf-8")
    request.cls.avatar_original_sha1 = hashlib.sha1(avatar_path.read_bytes()).hexdigest()

    with patch("slidge.db.avatar.avatar_cache.http", create=True) as mock:
        mock.get = mock_get
        mock.head = mock_get
        yield request


# just to have typings for the fixture which pycharm does not understand
class AvatarFixtureMixin:
    avatar_path: Path
    avatar_image: Image
    avatar_bytes: bytes
    avatar_sha1: str
    avatar_original_sha1: str
    avatar_url: str
    avatar_url_no_headers: str
    avatar_base64: str
    switch_avatar: Callable


class MockUUID4:
    def __init__(self, i: int) -> None:
        self.i = i
        self.hex = str(i)

    def __repr__(self):
        return self.i.__repr__()


class UUIDFixtureMixin(unittest.TestCase):
    def _next_uuid(self):
        self._uuid_counter += 1
        return self._uuid_counter

    def setUp(self):
        super().setUp()
        self._uuid_counter = 0
        self._uuid_patch = unittest.mock.patch("uuid.uuid4", side_effect=self._next_uuid)
        self._uuid_patch.start()

    def tearDown(self) -> None:
        super().tearDown()
        self._uuid_patch.stop()



import asyncio
import logging
from collections.abc import AsyncIterator, Hashable
from contextlib import asynccontextmanager
from typing import Any


class NamedLockMixin:
    def __init__(self, *a: Any, **k: Any) -> None:
        super().__init__(*a, **k)
        self.__locks = dict[Hashable, asyncio.Lock]()

    @asynccontextmanager
    async def lock(self, id_: Hashable) -> AsyncIterator[None]:
        log.trace("getting %s", id_)  # type:ignore
        locks = self.__locks
        if not locks.get(id_):
            locks[id_] = asyncio.Lock()
        try:
            async with locks[id_]:
                log.trace("acquired %s", id_)  # type:ignore
                yield
        finally:
            log.trace("releasing %s", id_)  # type:ignore
            waiters = locks[id_]._waiters
            if not waiters:
                del locks[id_]
                log.trace("erasing %s", id_)  # type:ignore

    def get_lock(self, id_: Hashable) -> asyncio.Lock | None:
        return self.__locks.get(id_)


log = logging.getLogger(__name__)



from .admin import MucAdminMixin
from .mam import MamMixin
from .misc import MucMiscMixin
from .owner import MucOwnerMixin
from .ping import PingMixin


class MucMixin(PingMixin, MamMixin, MucAdminMixin, MucOwnerMixin, MucMiscMixin):
    __slots__: list[str] = []


__all__ = ("MucMixin",)



from slixmpp.plugins.base import BasePlugin, register_plugin
from slixmpp.plugins.xep_0292.stanza import NS


class VCard4Provider(BasePlugin):
    name = "xep_0292_provider"
    description = "VCard4 Provider"
    dependencies = {"xep_0030"}

    def plugin_init(self) -> None:
        self.xmpp.plugin["xep_0030"].add_feature(NS)


register_plugin(VCard4Provider)



import logging
import string
import uuid
import warnings
from copy import copy
from datetime import datetime
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree as ET

import sqlalchemy as sa
from slixmpp import JID, InvalidJID, Message, Presence
from slixmpp.plugins.xep_0045.stanza import MUCAdminItem
from slixmpp.types import MessageTypes, OptJid
from sqlalchemy.orm.exc import DetachedInstanceError

from ..contact import LegacyContact
from ..core.mixins import ChatterDiscoMixin, MessageMixin, PresenceMixin
from ..core.mixins.db import DBMixin
from ..db.models import Participant
from ..util import SubclassableOnce, strip_illegal_chars
from ..util.types import (
    CachedPresence,
    Hat,
    LegacyMessageType,
    LegacyThreadType,
    MessageOrPresenceTypeVar,
    MucAffiliation,
    MucRole,
)

if TYPE_CHECKING:
    from .room import LegacyMUC


def strip_non_printable(nickname: str):
    new = (
        "".join(x for x in nickname if x in string.printable)
        + f"-slidge-{hash(nickname)}"
    )
    warnings.warn(f"Could not use {nickname} as a nickname, using {new}")
    return new


class LegacyParticipant(
    PresenceMixin,
    MessageMixin,
    ChatterDiscoMixin,
    DBMixin,
    SubclassableOnce,
):
    """
    A legacy participant of a legacy group chat.
    """

    is_participant = True

    mtype: MessageTypes = "groupchat"
    _can_send_carbon = False
    USE_STANZA_ID = True
    STRIP_SHORT_DELAY = False
    stored: Participant
    contact: LegacyContact[Any] | None

    def __init__(
        self,
        muc: "LegacyMUC",
        stored: Participant,
        is_system: bool = False,
        contact: LegacyContact[Any] | None = None,
    ) -> None:
        self.muc = muc
        self.session = muc.session
        self.xmpp = muc.session.xmpp
        self.is_system = is_system

        if contact is None and stored.contact is not None:
            contact = self.session.contacts.from_store(stored=stored.contact)
        if contact is not None and stored.contact is None:
            stored.contact = contact.stored

        self.stored = stored
        self.contact = contact

        super().__init__()

        if stored.resource is None:
            self.__update_resource(stored.nickname)

        self.log = logging.getLogger(f"{self.user_jid.bare}:{self.jid}")

    @property
    def is_user(self) -> bool:
        try:
            return self.stored.is_user
        except DetachedInstanceError:
            self.merge()
            return self.stored.is_user

    @is_user.setter
    def is_user(self, is_user: bool) -> None:
        with self.xmpp.store.session(expire_on_commit=True) as orm:
            orm.add(self.stored)
            self.stored.is_user = is_user
            orm.commit()

    @property
    def jid(self) -> JID:
        jid = JID(self.muc.jid)
        if self.stored.resource:
            jid.resource = self.stored.resource
        return jid

    @jid.setter
    def jid(self, x: JID):
        # FIXME: without this, mypy yields
        #        "Cannot override writeable attribute with read-only property"
        #        But it does not happen for LegacyContact. WTF?
        raise RuntimeError

    def __should_commit(self) -> bool:
        if self.is_system:
            return False
        if self.muc.get_lock("fill participants"):
            return False
        if self.muc.get_lock("fill history"):
            return False
        return True

    def commit(self, *args, **kwargs) -> None:
        if not self.__should_commit():
            return
        super().commit(*args, **kwargs)

    @property
    def user_jid(self):
        return self.session.user_jid

    def __repr__(self) -> str:
        return f"<Participant '{self.nickname}'/'{self.jid}' of '{self.muc}'>"

    @property
    def _presence_sent(self) -> bool:
        # we track if we already sent a presence for this participant.
        # if we didn't, we send it before the first message.
        # this way, event in plugins that don't map "user has joined" events,
        # we send a "join"-presence from the participant before the first message
        return self.stored.presence_sent

    @_presence_sent.setter
    def _presence_sent(self, val: bool) -> None:
        if self._presence_sent == val:
            return
        self.stored.presence_sent = val
        if not self.__should_commit():
            return
        with self.xmpp.store.session() as orm:
            orm.execute(
                sa.update(Participant)
                .where(Participant.id == self.stored.id)
                .values(presence_sent=val)
            )
            orm.commit()

    @property
    def nickname_no_illegal(self) -> str:
        return self.stored.nickname_no_illegal

    @property
    def affiliation(self):
        return self.stored.affiliation

    @affiliation.setter
    def affiliation(self, affiliation: MucAffiliation) -> None:
        if self.affiliation == affiliation:
            return
        self.stored.affiliation = affiliation
        if not self.muc.participants_filled:
            return
        self.commit()
        if not self._presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    def send_affiliation_change(self) -> None:
        # internal use by slidge
        msg = self._make_message()
        msg["muc"]["affiliation"] = self.affiliation
        msg["type"] = "normal"
        if not self.muc.is_anonymous and not self.is_system:
            if self.contact:
                msg["muc"]["jid"] = self.contact.jid
            else:
                warnings.warn(
                    f"Private group but no 1:1 JID associated to '{self}'",
                )
        self._send(msg)

    @property
    def role(self):
        return self.stored.role

    @role.setter
    def role(self, role: MucRole) -> None:
        if self.role == role:
            return
        self.stored.role = role
        if not self.muc.participants_filled:
            return
        self.commit()
        if not self._presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    @property
    def hats(self) -> list[Hat]:
        return [Hat(*h) for h in self.stored.hats] if self.stored.hats else []

    def set_hats(self, hats: list[Hat]) -> None:
        if self.hats == hats:
            return
        self.stored.hats = hats  # type:ignore[assignment]
        if not self.muc.participants_filled:
            return
        self.commit(merge=True)
        if not self._presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    def __update_resource(self, unescaped_nickname: str | None) -> None:
        if not unescaped_nickname:
            self.stored.resource = ""
            if self.is_system:
                self.stored.nickname_no_illegal = ""
            else:
                warnings.warn(
                    "Only the system participant is allowed to not have a nickname"
                )
                nickname = f"unnamed-{uuid.uuid4()}"
                self.stored.resource = self.stored.nickname_no_illegal = nickname
            return

        self.stored.nickname_no_illegal, jid = escape_nickname(
            self.muc.jid,
            unescaped_nickname,
        )
        self.stored.resource = jid.resource

    def send_configuration_change(self, codes: tuple[int]):
        if not self.is_system:
            raise RuntimeError("This is only possible for the system participant")
        msg = self._make_message()
        msg["muc"]["status_codes"] = codes
        self._send(msg)

    @property
    def nickname(self):
        return self.stored.nickname

    @nickname.setter
    def nickname(self, new_nickname: str) -> None:
        old = self.nickname
        if new_nickname == old:
            return

        if self.muc.stored.id is not None:
            with self.xmpp.store.session() as orm:
                if not self.xmpp.store.rooms.nick_available(
                    orm, self.muc.stored.id, new_nickname
                ):
                    if self.contact is None:
                        new_nickname = f"{new_nickname} ({self.occupant_id})"
                    else:
                        new_nickname = f"{new_nickname} ({self.contact.legacy_id})"

        cache = getattr(self, "_last_presence", None)
        if cache:
            last_seen = cache.last_seen
            kwargs = cache.presence_kwargs
        else:
            last_seen = None
            kwargs = {}

        kwargs["status_codes"] = {303}

        p = self._make_presence(ptype="unavailable", last_seen=last_seen, **kwargs)
        # in this order so pfrom=old resource and we actually use the escaped nick
        # in the muc/item/nick element
        self.__update_resource(new_nickname)
        p["muc"]["item"]["nick"] = self.jid.resource
        self._send(p)

        self.stored.nickname = new_nickname
        self.commit()
        kwargs["status_codes"] = set()
        p = self._make_presence(ptype="available", last_seen=last_seen, **kwargs)
        self._send(p)

    def _make_presence(
        self,
        *,
        last_seen: datetime | None = None,
        status_codes: set[int] | None = None,
        user_full_jid: JID | None = None,
        **presence_kwargs,
    ):
        p = super()._make_presence(last_seen=last_seen, **presence_kwargs)
        p["muc"]["affiliation"] = self.affiliation
        p["muc"]["role"] = self.role
        if self.hats:
            p["hats"].add_hats(self.hats)
        codes = status_codes or set()
        if self.is_user:
            codes.add(110)
        if not self.muc.is_anonymous and not self.is_system:
            if self.is_user:
                if user_full_jid:
                    p["muc"]["jid"] = user_full_jid
                else:
                    jid = JID(self.user_jid)
                    try:
                        jid.resource = next(iter(self.muc.get_user_resources()))
                    except StopIteration:
                        jid.resource = "pseudo-resource"
                    p["muc"]["jid"] = self.user_jid
                codes.add(100)
            elif self.contact:
                p["muc"]["jid"] = self.contact.jid
                if a := self.contact.get_avatar():
                    p["vcard_temp_update"]["photo"] = a.id
            else:
                warnings.warn(
                    f"Private group but no 1:1 JID associated to '{self}'",
                )
        if self.is_user and (hash_ := self.session.user.avatar_hash):
            p["vcard_temp_update"]["photo"] = hash_
        p["muc"]["status_codes"] = codes
        return p

    @property
    def DISCO_NAME(self):  # type:ignore[override]
        return self.nickname

    def __send_presence_if_needed(
        self, stanza: Message | Presence, full_jid: JID, archive_only: bool
    ) -> None:
        if (
            archive_only
            or self.is_system
            or self.is_user
            or self._presence_sent
            or stanza["subject"]
        ):
            return
        if isinstance(stanza, Message):
            if stanza.get_plugin("muc", check=True):
                return
            self.send_initial_presence(full_jid)

    @property
    def occupant_id(self) -> str:
        return self.stored.occupant_id

    def _send(
        self,
        stanza: MessageOrPresenceTypeVar,
        full_jid: JID | None = None,
        archive_only: bool = False,
        legacy_msg_id=None,
        initial_presence=False,
        **send_kwargs,
    ) -> MessageOrPresenceTypeVar:
        if stanza.get_from().resource:
            stanza["occupant-id"]["id"] = self.occupant_id
        else:
            stanza["occupant-id"]["id"] = "room"
        self.__add_nick_element(stanza)
        if not self.is_user and isinstance(stanza, Presence):
            if stanza["type"] == "unavailable" and not self._presence_sent:
                return stanza  # type:ignore
            if initial_presence:
                self.stored.presence_sent = True
            else:
                self._presence_sent = True
        if full_jid:
            stanza["to"] = full_jid
            self.__send_presence_if_needed(stanza, full_jid, archive_only)
            if self.is_user:
                assert stanza.stream is not None
                stanza.stream.send(stanza, use_filters=False)
            else:
                stanza.send()
        else:
            if hasattr(self.muc, "archive") and isinstance(stanza, Message):
                self.muc.archive.add(stanza, self, archive_only, legacy_msg_id)
            if archive_only:
                return stanza
            for user_full_jid in self.muc.user_full_jids():
                stanza = copy(stanza)
                stanza["to"] = user_full_jid
                self.__send_presence_if_needed(stanza, user_full_jid, archive_only)
                stanza.send()
        return stanza

    def mucadmin_item(self):
        item = MUCAdminItem()
        item["nick"] = self.nickname
        item["affiliation"] = self.affiliation
        item["role"] = self.role
        if not self.muc.is_anonymous:
            if self.is_user:
                item["jid"] = self.user_jid.bare
            elif self.contact:
                item["jid"] = self.contact.jid.bare
            else:
                warnings.warn(
                    (
                        f"Private group but no contact JID associated to {self.jid} in"
                        f" {self}"
                    ),
                )
        return item

    def __add_nick_element(self, stanza: Presence | Message) -> None:
        if (nick := self.nickname_no_illegal) != self.jid.resource:
            n = self.xmpp.plugin["xep_0172"].stanza.UserNick()
            n["nick"] = nick
            stanza.append(n)

    def _get_last_presence(self) -> CachedPresence | None:
        own = super()._get_last_presence()
        if own is None and self.contact:
            return self.contact._get_last_presence()
        return own

    def send_initial_presence(
        self,
        full_jid: JID,
        nick_change: bool = False,
        presence_id: str | None = None,
    ) -> None:
        """
        Called when the user joins a MUC, as a mechanism
        to indicate to the joining XMPP client the list of "participants".

        Can be called this to trigger a "participant has joined the group" event.

        :param full_jid: Set this to only send to a specific user XMPP resource.
        :param nick_change: Used when the user joins and the MUC renames them (code 210)
        :param presence_id: set the presence ID. used internally by slidge
        """
        #  MUC status codes: https://xmpp.org/extensions/xep-0045.html#registrar-statuscodes
        codes = set()
        if nick_change:
            codes.add(210)

        if self.is_user:
            # the "initial presence" of the user has to be vanilla, as it is
            # a crucial part of the MUC join sequence for XMPP clients.
            kwargs = {}
        else:
            cache = self._get_last_presence()
            self.log.debug("Join muc, initial presence: %s", cache)
            if cache:
                ptype = cache.ptype
                if ptype == "unavailable":
                    return
                kwargs = dict(
                    last_seen=cache.last_seen, pstatus=cache.pstatus, pshow=cache.pshow
                )
            else:
                kwargs = {}
        p = self._make_presence(
            status_codes=codes,
            user_full_jid=full_jid,
            **kwargs,  # type:ignore
        )
        if presence_id:
            p["id"] = presence_id
        self._send(p, full_jid, initial_presence=True)

    def leave(self) -> None:
        """
        Call this when the participant leaves the room
        """
        self.muc.remove_participant(self)

    def kick(self, reason: str | None = None) -> None:
        """
        Call this when the participant is kicked from the room
        """
        self.muc.remove_participant(self, kick=True, reason=reason)

    def ban(self, reason: str | None = None) -> None:
        """
        Call this when the participant is banned from the room
        """
        self.muc.remove_participant(self, ban=True, reason=reason)

    def get_disco_info(self, jid: OptJid = None, node: str | None = None):
        if self.contact is not None:
            return self.contact.get_disco_info()
        return super().get_disco_info()

    def moderate(
        self, legacy_msg_id: LegacyMessageType, reason: str | None = None
    ) -> None:
        for i in self._legacy_to_xmpp(legacy_msg_id):
            m = self.muc.get_system_participant()._make_message()
            m["retract"]["id"] = i
            if self.is_system:
                m["retract"].enable("moderated")
            else:
                m["retract"]["moderated"]["by"] = self.jid
                m["retract"]["moderated"]["occupant-id"]["id"] = self.occupant_id
            if reason:
                m["retract"]["reason"] = reason
            self._send(m)

    def set_room_subject(
        self,
        subject: str,
        full_jid: JID | None = None,
        when: datetime | None = None,
        update_muc: bool = True,
    ) -> None:
        if update_muc:
            self.muc._subject = subject  # type: ignore
            self.muc.subject_setter = self.nickname
            self.muc.subject_date = when

        msg = self._make_message()
        if when is not None:
            msg["delay"].set_stamp(when)
            msg["delay"]["from"] = self.muc.jid
        if subject:
            msg["subject"] = subject
        else:
            # may be simplified if slixmpp lets it do it more easily some day
            msg.xml.append(ET.Element(f"{{{msg.namespace}}}subject"))
        self._send(msg, full_jid)

    def set_thread_subject(
        self,
        thread: LegacyThreadType,
        subject: str | None,
        when: datetime | None = None,
    ) -> None:
        msg = self._make_message()
        msg["thread"] = str(thread)
        if when is not None:
            msg["delay"].set_stamp(when)
            msg["delay"]["from"] = self.muc.jid
        if subject:
            msg["subject"] = subject
        else:
            # may be simplified if slixmpp lets it do it more easily some day
            msg.xml.append(ET.Element(f"{{{msg.namespace}}}subject"))
        self._send(msg)


def escape_nickname(muc_jid: JID, nickname: str) -> tuple[str, JID]:
    nickname = nickname_no_illegal = strip_illegal_chars(nickname).replace("\n", " | ")

    jid = JID(muc_jid)

    try:
        jid.resource = nickname
    except InvalidJID:
        nickname = nickname.encode("punycode").decode()
        try:
            jid.resource = nickname
        except InvalidJID:
            # at this point there still might be control chars
            jid.resource = strip_non_printable(nickname)

    return nickname_no_illegal, jid


log = logging.getLogger(__name__)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from ...db import GatewayUser
from .. import config
from .util import DispatcherMixin

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway


class RegistrationMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: BaseGateway) -> None:
        super().__init__(xmpp)
        xmpp["xep_0077"].api.register(
            self.xmpp.make_registration_form, "make_registration_form"
        )
        xmpp["xep_0077"].api.register(self._user_get, "user_get")
        xmpp["xep_0077"].api.register(self._user_validate, "user_validate")
        xmpp["xep_0077"].api.register(self._user_modify, "user_modify")
        # kept for slixmpp internal API compat
        # TODO: either fully use slixmpp internal API or rewrite registration without it at all
        xmpp["xep_0077"].api.register(lambda *a: None, "user_remove")

        xmpp.add_event_handler("user_register", self._on_user_register)
        xmpp.add_event_handler("user_unregister", self._on_user_unregister)

    def get_user(self, jid: JID) -> GatewayUser | None:
        session = self.xmpp.get_session_from_jid(jid)
        if session is None:
            return None
        return session.user

    async def _user_get(
        self, _gateway_jid, _node, ifrom: JID, iq: Iq
    ) -> GatewayUser | None:
        if ifrom is None:
            ifrom = iq.get_from()
        return self.get_user(ifrom)

    async def _user_validate(self, _gateway_jid, _node, ifrom: JID, iq: Iq) -> None:
        xmpp = self.xmpp
        log.debug("User validate: %s", ifrom.bare)
        form_dict = {f.var: iq.get(f.var) for f in xmpp.REGISTRATION_FIELDS}
        xmpp.raise_if_not_allowed_jid(ifrom)
        legacy_module_data = await xmpp.user_prevalidate(ifrom, form_dict)
        if legacy_module_data is None:
            legacy_module_data = form_dict
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=ifrom.bare,
                legacy_module_data=legacy_module_data,
            )
            orm.add(user)
            orm.commit()
            log.info("New user: %s", user)

    async def _user_modify(
        self, _gateway_jid, _node, ifrom: JID, form_dict: dict[str, str | None]
    ):
        await self.xmpp.user_prevalidate(ifrom, form_dict)
        log.debug("Modify user: %s", ifrom)
        with self.xmpp.store.session() as orm:
            user = orm.query(GatewayUser).one_or_none()
        if user is None:
            raise XMPPError("internal-server-error", "User not found")
        user.legacy_module_data.update(form_dict)
        self.xmpp.store.users.update(user)

    async def _on_user_register(self, iq: Iq) -> None:
        session = await self._get_session(iq, wait_for_ready=False)
        for jid in config.ADMINS:
            self.xmpp.send_message(
                mto=jid,
                mbody=f"{iq.get_from()} has registered",
                mtype="chat",
                mfrom=self.xmpp.boundjid.bare,
            )
        session.send_gateway_message(self.xmpp.WELCOME_MESSAGE)
        await self.xmpp.login_wrap(session)

    async def _on_user_unregister(self, iq: Iq) -> None:
        await self.xmpp.kill_session(iq.get_from())


log = logging.getLogger(__name__)



# This module contains patches for slixmpp; some have pending requests upstream
# and should be removed on the next slixmpp release.
import uuid

# ruff: noqa: F401
import slixmpp.plugins
import slixmpp.stanza.roster
from slixmpp import Iq, Message
from slixmpp.exceptions import IqError
from slixmpp.plugins.xep_0050 import XEP_0050, Command
from slixmpp.plugins.xep_0356.permissions import IqPermission
from slixmpp.plugins.xep_0356.privilege import XEP_0356, PrivilegedIqError
from slixmpp.plugins.xep_0469.stanza import NS as PINNED_NS
from slixmpp.plugins.xep_0469.stanza import Pinned
from slixmpp.xmlstream import StanzaBase

from ..util.archive_msg import set_client_namespace
from . import (
    link_preview,
    xep_0077,
    xep_0100,
    xep_0153,
    xep_0292,
)


def set_pinned(self, val: bool) -> None:
    extensions = self.parent()
    if val:
        extensions.enable("pinned")
    else:
        extensions._del_sub(f"{{{PINNED_NS}}}pinned")


Pinned.set_pinned = set_pinned


def session_bind(self, jid) -> None:
    self.xmpp["xep_0030"].add_feature(Command.namespace)
    # awful hack to for the disco items: we need to comment this line
    # related issue: https://todo.sr.ht/~nicoco/slidge/131
    # self.xmpp['xep_0030'].set_items(node=Command.namespace, items=tuple())


XEP_0050.session_bind = session_bind  # type:ignore


def reply(self, body=None, clear: bool = True):
    """
    Overrides slixmpp's Message.reply(), since it strips to sender's resource
    for mtype=groupchat, and we do not want that, because when we raise an XMPPError,
    we actually want to preserve the resource.
    (this is called in RootStanza.exception() to handle XMPPErrors)
    """
    new_message = StanzaBase.reply(self, clear)
    new_message["thread"] = self["thread"]
    new_message["parent_thread"] = self["parent_thread"]

    del new_message["id"]
    if self.stream is not None and self.stream.use_message_ids:
        new_message["id"] = self.stream.new_id()

    if body is not None:
        new_message["body"] = body
    return new_message


async def send_privileged_iq(self, encapsulated_iq: Iq, iq_id: str | None = None) -> Iq:
    """
    Send an IQ on behalf of a user

    Caution: the IQ *must* have the jabber:client namespace

    Raises :class:`PrivilegedIqError` on failure.
    """
    iq_id = iq_id or str(uuid.uuid4())
    encapsulated_iq["id"] = iq_id
    if encapsulated_iq.namespace != "jabber:client":
        set_client_namespace(encapsulated_iq)
    server = encapsulated_iq.get_from().domain
    perms = self.granted_privileges.get(server)
    if not perms:
        raise PermissionError(f"{server} has not granted us any privilege")
    itype = encapsulated_iq["type"]
    for ns in encapsulated_iq.plugins.values():
        type_ = perms.iq[ns.namespace]
        if type_ == IqPermission.NONE:
            raise PermissionError(
                f"{server} has not granted any IQ privilege for namespace {ns.namespace}"
            )
        elif type_ == IqPermission.BOTH:
            pass
        elif type_ != itype:
            raise PermissionError(
                f"{server} has not granted IQ {itype} privilege for namespace {ns.namespace}"
            )
    iq = self.xmpp.make_iq(
        itype=itype,
        ifrom=self.xmpp.boundjid.bare,
        ito=encapsulated_iq.get_from(),
        id=iq_id,
    )
    iq["privileged_iq"].append(encapsulated_iq)

    try:
        resp = await iq.send()
    except IqError as exc:
        raise PrivilegedIqError(exc.iq)

    return resp["privilege"]["forwarded"]["iq"]


XEP_0356.send_privileged_iq = send_privileged_iq  # type:ignore
Message.reply = reply  # type: ignore

slixmpp.plugins.PLUGINS.extend(
    [
        "link_preview",
        "xep_0292_provider",
    ]
)
